"""
CoRNy Basics
    General helper functions that make life easier
"""

from corny import *

name_std = '_std'
name_err = '_err'

def debug():
    """
        Sets a marker at this position where the running script stops
        and enters interactive mode for debugging
    """
    import pdb
    pdb.set_trace()

def one_of(a, b, default=np.nan):
    """
    Return a or b or default.
    E.g.: latitude = one_of(lat_column, station_lat)
    """
    if a:
        return(a)
    elif b:
        return(b)
    else:
        return(default)

def safe_cast(var, to_type, default=None):
    """
    Convert variable type (int, float, str, ...) safely.
    Use:
    """
    if default is None:
         if   to_type==float: default = np.nan
         elif to_type==str  : default = ''

    try:
        return(to_type(var))
    except (ValueError, TypeError):
        return(default)


def col_is_unique(s):
    """
    Check whether a pandas column has all identical values.
    Usage:
        if col_is_unique(data.lon): data = data.iloc[0]
    """
    import numpy
    if isinstance(s, numpy.ndarray):
        a = s
    else:
        import shapely
        if isinstance(s[0], shapely.geometry.point.Point):
            a = np.array([pg.x for pg in s])
        else:
            import pandas
            if isinstance(s, pandas.Series):
                a = s.to_numpy()
            else:
                print('Err','col_is_unique: variable is neither of ndarray, pandas series, shapely point.')

    # Boolean array
    b = a[0] == a
    # Check all
    return(b.all())


def str2dt(s, tz=None):
    """
    Parses the date time string as a datetime object
    Usage:
        str2dt('2021-12-24', tz='UTC')
        = datetime object
    """
    dto = dateutil.parser.parse(s) # uses dateutil
    if tz is None:
        return(dto)
    elif tz == 'UTC':
        return(dto.replace(tzinfo=pytz.UTC))
    else:
        return(dto.replace(tzinfo=tz))

def interpret_time_format(s):
    s = s.replace('monthname', '%b').replace('month', '%M')
    s = s.replace('day', '%d').replace('hour', '%H').replace('year', '%Y')
    s = s.replace('minute', '%M').replace('min', '%M')
    return(s)

def time_resolution(data, unit='sec'):
    """
    Estimate time resolution of pandas in given units.
    """
    tres = data.index.to_series().dropna().diff().median().total_seconds()

    if   unit == 'min':  tres *= 60
    elif unit == 'hour': tres *= 60*60
    elif unit == 'day':  tres *= 60*60*24

    return(tres)


def clean_interval_str(a):
    """
    return valid interval string (1Min, 2H, ...) from user input
    """

    # if number is provided, assume it's in hours
    if a.isdigit(): a = str(a)+'hour'

    # harmonize spelling
    a = a.replace('sec','S').replace('hour','H').replace('min','Min').replace('day','D').replace('week','W').replace('month','M').replace('year','A')

    # remove plural s
    a = a.replace('s','')

    return(a)


def aggregate(data, aggstr, func='mean', set_std=[], adapt_err=[], min_count=None, verbose=False):
    """
    Aggregation of the whole data set.
    Adds stddev columns for specific columns.
    Adapts error by 1/sqrt(a) for specific columns.
    Usage:
        aggregate(data, '1H')
        = pandas
    """

    # Clean aggregation window string
    aggstr = clean_interval_str(aggstr)

    if verbose: print('| Aggregate data to ' + aggstr + ' time steps...', end='')

    # record time resolution
    tres_before = time_resolution(data)

    # Resample
    aggobj = data.resample(aggstr)
    if   func == 'mean':   data = aggobj.mean()
    elif func == 'median': data = aggobj.median()
    elif func == 'sum':    data = aggobj.sum(min_count=min_count)
    else:
        print('Err','Aggregation method unknown: ' + method)

    # record time resolution
    tres_after = time_resolution(data)

    # factor by which the time resolution is scaled up
    tres_factor = tres_after / tres_before
    tres_factor_err = 1 / np.sqrt( tres_factor )

    if verbose: print(' (data x%0.2f, error x%.02f)' % (tres_factor, tres_factor_err))

    # Errors scale with sqrt(1/tresfactor)
    for column in adapt_err:
        data[column + name_err] *= tres_factor_err

    # Add empirical standard deviation
    for column in set_std:
        data[column + name_std] = aggobj.std()[column]

    return(data)


def minly(data, agg=1, **kwargs):
    aggstr = '%iMin' % agg
    data = aggregate(data, aggstr, **kwargs)
    return(data)

def hourly(data, agg=1, **kwargs):
    aggstr = '%iH' % agg
    data = aggregate(data, aggstr, **kwargs)
    return(data)

def daily(data, agg=1, **kwargs):
    aggstr = '%iD' % agg
    data = aggregate(data, aggstr, **kwargs)
    data.index.freq = 'D'
    return(data)


# Moving average
def moving_avg(D, column, window=1, err=True, std=True, center=True):
    window = int(window)
    suffix = '_mov' #+str(window)
    rollobj = D[column].rolling(window, center=center)
    D[column+suffix]                = rollobj.mean()
    if std: D[column+suffix+'_std'] = rollobj.std()
    if err: D[column+suffix+'_err'] = D[column+'_err'] * 1/np.sqrt(window)
    return(D)


def merge_datasets(data_1, data_2, name_1='', name_2=''):
    """
    Merge columns of two DataFrames by index
    """

    with Progress('  Merging dataset %6s into %6s     ' % (name_2, name_1)) as Pr:
        data = pandas.merge_asof(data_1, data_2, right_index=True, left_index=True)
        Pr.update('(%5d lines x %2d columns)' % (len(data),len(data.columns)))

    return(data)


def cut_period_to(D, a, b=None):
    """
    Cut period to that of another data frame.
    Usage:
        cut_period(df, '2019-03-14', '2020-03-14')
        cut_period(df, datetime, datetime)
        cut_period(df, shorter_df)
        = df
    """
    with Progress('Cutting period') as Pr:

        if a is None or a.strip() == '':
            a = D.index.min()
        if b is None or b.strip() == '':
            b = D.index.max()

        # Cut to other DataFrame
        if isinstance(a, pandas.DataFrame):
            a = a.index.min()
            b = a.index.max()

        # Strings to parse
        elif isinstance(a, str):
            if a.strip() == 'today':
                a = datetime.now().strftime('%Y-%m-%d')
                b = D.index.max().strftime('%Y-%m-%d %H:%M')

            a = str2dt(a, 'UTC')
            if isinstance(b, str):
                b = str2dt(b, 'UTC')

        # Datetimes
        Pr.update('(%s to %s)' % (a,b))
        return(D.loc[a:b])

def new_column(data, copy_from=None, fill=np.nan):
    """
    Create new column from a copy of another column.
    If no other or no valid column is given, fill with fill.
    If no fill is given, do nothing.
    Example #1:
        data['lat'] = new_column(data, copy_from='Latitude')
        # Does nothing if Latitude is not in data
    Example #2:
        data['pressure'] = new_column(data, copy_from='p4', fill=np.nan)
        # Creates a new column filled with np.nan even if no p4 exists

    """
    if copy_from:
        if copy_from in data.columns:
            return(data[copy_from].copy())
        else:
            print('Warn','Cannot find column "%s" in the dataset.' % copy_from)

    return(fill)


def Performance(data, title='Test', writeout=True):
    """
    Evaluation of two columns of a df
    Usage:
        Performance(data[[a,b]]
        = [kge, nse, rmse, correl, var, bias]
    """
    from hydroeval import evaluator, kge, rmse, nse

    columns = data.columns
    D = data.copy().dropna().values
    k = evaluator(kge,  D[:,0], D[:,1]).flatten()
    r = evaluator(rmse, D[:,0], D[:,1])
    n = evaluator(nse,  D[:,0], D[:,1])
    if writeout:
        print('%35s: KGE %.3f, NSE %.3f, RMSE %.3f, Correlation %.3f, Variance %.3f, Bias %.3f' %
            (columns[0]+' vs. '+columns[1], k[0], n[0], r[0], k[1], k[2], k[3]))
    else:
        return([k[0], n[0], r[0], k[1], k[2], k[3]])

def optimize(df, opt, cmp, span=(700,1100), func='kge'):
    if   func=='kge':
        fi = 0
        best = -1
    elif func=='nse':
        fi = 1
        best = -1
    elif func=='rmse':
        fi = 2
        best = 100000
    else: print('Err','Unknown function. Choose '+'kge, NSE, RMSE')

    best_i = 0
    for i in np.arange(span[0],span[1]):
        df[opt+'_opt'] = df[opt] * i
        p = Performance(df[[opt+'_opt', cmp]], writeout=False)
        if fi == 2:
            if p[fi] < best:
                best = p[fi]
                best_i = i
        else:
            if p[fi] > best:
                best = p[fi]
                best_i = i

    df[opt+'_opt'] = df[opt] * best_i
    print("%5i" % best_i, end='')
    Performance(df[[opt+'_opt', cmp]])


def ifdef(a, b, convert=float):
    if not a:
        return(b)
    else:
        return(convert(a))

def xsplit(s, range=0, type=str):
    """
    Split a string by , and ignore spaces,
    convert to certain data type,
    optionally generate an array of length `range`.
    Example:
        coords = xplit('51.5, 12.1, 120', range=3, type=float)
    """
    if range is None or range == []:
        range = 0
        
    if isinstance(type, str):
        typedict = {'str':str,'float':float,'int':int,'bool':bool}
        type = typedict[type]

    if   type=='str':   type = str
    elif type=='int':   type = int
    elif type=='float': type = float
    elif type=='bool':  type = bool

    if isinstance(s, str):
        s = s.replace(',',' ')
        a = np.array(s.split(), dtype=type)
    else:
        a = np.array([])

    # fill up with NaNs
    for i in np.arange(range-len(a)):
        a = np.append(a, np.nan)
    return(a)

# Split a string by \s,\s into an array
def csplit(s, range=0, type=str):
    s = re.sub(r'\s*,+\s*', ',', s)
    s = re.sub(r',+', ',', s)
    a = np.array(s.split(','), dtype=type)
    return(a)

def chomp(x):
    if x.endswith("\r\n"): return x[:-2]
    if x.endswith("\n") or x.endswith("\r"): return x[:-1]
    return(x)

# Spatial calculations

def latlon2utm(lats, lons):
    import pyproj
    if isinstance(lats, pandas.DataFrame) or isinstance(lats, pandas.Series):
        lats = lats.values
        lons = lons.values

    ## Deprecated:
    # import pyproj
    # utmy, utmx = pyproj.transform(pyproj.Proj("epsg:4326"), pyproj.Proj("epsg:31468"), lats, lons)
    ## New pyproj2 syntax:
    from pyproj import Transformer
    utmy, utmx = Transformer.from_crs("epsg:4326", "epsg:31468").transform(lats, lons)

    return(utmy, utmx)

def utm2latlon(utmx, utmy):
    import pyproj
    if isinstance(utmx, pandas.DataFrame) or isinstance(utmx, pandas.Series):
        utmx = utmx.values
        utmy = utmy.values

    ## Deprecated:
    # import pyproj
    # lats, lons = pyproj.transform(pyproj.Proj("epsg:31468"), pyproj.Proj("epsg:4326"), utmy, utmx)
    ## New pyproj2 syntax:
    from pyproj import Transformer
    lats, lons = Transformer.from_crs("epsg:31468", "epsg:4326").transform(utmy, utmx)

    return(lats, lons)


def fineround(x, reso=3, fine=1):
    # A rounding function that can round to values even in between digits
    # e.g. use fine=1 to round to the nearest .0 (standard digit round)
    #          fine=2 to round to the nearest .0 or .5
    #          ...

    return(np.round(x*fine, int(reso))/fine)


#######################################
#   Spatial Geoanalysis               #
#######################################


#JJ recalculate strange coordinates to proper decimals
# argument is a pandas.DataFrame(columns=[number, direction])
# remember that C.iloc[:,i] is the i^th column of C

def deg100min2dec(C):
    # Convert
    x = C.iloc[:,0] / 100
    deg = np.floor(x)
    dec = deg + (x - deg)/0.6
    # Negate if cardinal direction is south or west
    dec *= C.iloc[:,1].map({'N': 1, 'S': -1, 'E': 1, 'W': -1})
    return(dec)


#JJ use x nearest neighbors for averaging
## find k closest lat/lon to every lat/lon coordinates
def distance(pointA, pointB):
    return sqrt((pointB[0] - pointA[0]) ** 2 + (pointB[1] - pointA[1]) ** 2)

import zipfile
import os


def distanceinkm(xstand,ystand,xfor,yfor): # approximate radius of earth in km

    R = 6373.0
    lat1 = radians(xstand)
    lon1 = radians(ystand)
    lat2 = radians(xfor)
    lon2 = radians(yfor)

    dlon = lon2 - lon1
    dlat = lat2 - lat1
    a = sin(dlat / 2)**2 + cos(lat1) * cos(lat2) * sin(dlon / 2)**2
    c = 2 * atan2(sqrt(a), sqrt(1 - a))
    distance = R * c
    return(distance)

def Neighbours(lat, lon, numpoints, data):
    numpoints = numpoints+1 # points itself included - therefore add one to get desired output
    points = np.vstack((lat, lon)).T # create numpy array from coordinates

    kdt = KDTree(points, leaf_size=30, metric='euclidean') # build tree with euclidean distances
    neighborpoints = kdt.query(points, k=numpoints, return_distance=False) # find the closest k points

    # calculate maximal distance within the k clostest coordinates
    neighbordist = []
    for i in range(len(points)):
        furthest = np.array(nlargest(1, points[neighborpoints[i]], lambda p : distance(p, (points[i])))) # find the coordinates for the k closest points
        neighbordist.append(distanceinkm(points[i][0],points[i][1], furthest[0][0],furthest[0][1])*1000)

    return neighbordist, neighborpoints

    #neighborpoints = [] # create empty list for all results

    #for i in range(len(points)): ## iterate for all points
    #    nearest = np.array(nsmallest(numpoints, points, lambda p : distance(p, (points[i])))) # find the coordinates for the k closest points
    #    print(nearest)
    #    index = [] # create empty list for the index of the positions

    #    for m in range(len(nearest)):
    #        idx, rest = np.where(points==nearest[m]) # find the rows of these closest k coordinates
    #        idx = stats.mode(idx) # calculate the mode - necessary because sometimes several outputs from above
    #        index.append(idx.mode.tolist()) # create list from the k rows found

    #    neighborpoints.append(index) # create list of k closest locations for all locations
    #for i in range(len(neighborpoints)):
    #    neighborpoints[i] = list(itertools.chain(*neighborpoints[i])) # reduce for one nested list


#######################################
#   Console                           #
#######################################


def report(data, column, meanof=None, correction='', units='', format='i %s'):
    if meanof is None: meanof = column
    evallist = [column] + [data[c].mean() for c in meanof] + [units]
    s = format % tuple(evallist)
    if correction != '':
        if isinstance(correction, list):
            q = 1
            for c in correction:
                q *= data[c]
        elif correction in data.columns:
            q = data[correction]
        s += ' (%+.0f%%)' % (100*(1-q.mean()))
    print(cprintf(s, 'cyan'))

import os
def cprintf(s, fmt=''):
    os.system("") # makes it work also in the Win10 Console
    a = ''
    if fmt=='black':     a='\033[30m'
    if fmt=='red':       a='\033[31m'
    if fmt=='green':     a='\033[32m'
    if fmt=='yellow':    a='\033[33m'
    if fmt=='blue':      a='\033[34m'
    if fmt=='magenta':   a='\033[35m'
    if fmt=='cyan':      a='\033[36m'
    if fmt=='white':     a='\033[37m'
    if fmt=='underline': a='\033[4m'
    if a != '':
        return(a + s + '\033[0m')
    else:
        return(s)

"""
Progress environment
    Why?
        No more cluttering of print statements before and after code
    Usage:
        with Progress(text, format) as Pr:
            pass
    Example #1:
        with Progress('Calculating'):
            x = 1+1
        # Returns: | Calculating ... OK.
    Example #2:
        with Progress('Reading file',fmt='R') as Pr:
            try:
                open(file, 'r')
            except:
                Pr.fail()
        # Returns: > Reading file ... FAILED!
"""
import traceback

class Progress():
    def __init__(self, *args, fmt='Proc'):
        self._fail = False
        self.args = args
        self._fmt = fmt

    def print_status(self, s, end=None):
        fmt = self._fmt
        if fmt.strip()!='':
            fmt += '_only_color'
        print(fmt, ' '+s, end=end)

    def update(self, s):
        self.print_status(s+'...', end='')

    def fail(self):
        self._fail = True
        self.print_status('FAILED!')

    def __enter__(self):
        print(self._fmt, *self.args, '...', end='')
        return(self)

    def __exit__(self, except_type, except_value, tb):
        if not except_type:
            self.print_status('OK.')
        else:
            self.print_status('FAILED.')
            print('Err', traceback.format_exc())
            #traceback.print_exc()
        return(True)

class Progressbar():
    """
    Progressbar environment
    Usage:
        with Progressbar(size, title, length) as Pb:
            Pb.update(text, index)
    Example #1:
        L = ['a','b','c','d','e']
        with Progressbar(L) as Pb:
            for l in L:
                Pb.update()
        # Returns: | Please wait >>>>>>>               25%

    Example #2:
        files = glob('*.txt')
        with Progressbar(len(files), 'Reading') as Pb:
            i = 0
            for file in files:
                Pb.update(file, i)
                fh = open(file, 'r')
                i += 1
    # Returns: | Reading    >>>>>>>               25% readme.txt
    """

    def __init__(self, total, title='Please wait', length=None):

        # total counts
        if isinstance(total,int):
            self.total = total
        elif isinstance(total,float) or isinstance(total,str):
            try:
                self.total = int(total)
            except:
                print('Err','Progressbar cannot understand total bar length: ', total)
        elif isinstance(total,list):
            self.total = len(total)
        else:
            try:
                self.total = len(list(total))
            except:
                print('Err','Progressbar cannot understand total bar length: ', total)

        # Title
        self.title = title

        # Terminal length
        if length is None:
            import shutil
            self.length = shutil.get_terminal_size()[0]-79
        else:
            self.length = length

        self.comment = ''

    def update(self, comment='', progress=None):
        if progress is None:
            self.progress += 1
        else:
            self.progress = progress
        self.comment = comment
        self.printstr()

    def printstr(self):
        self.progress_pc = round((self.progress)/self.total*100, 1)
        progress_str = '>' * int(round(self.progress_pc/(100/self.length)))
        remaining_str = ' ' * int(self.length - len(progress_str))
        print('P','%-11s %s%s %3.0f%% %s' % (self.title, progress_str, remaining_str, self.progress_pc, self.comment), end='\r')
        sys.stdout.flush()

    def __enter__(self):
        self.update('',-1)
        return(self)

    def __exit__(self, type, value, traceback):
        self.update(' '*60, self.total)

"""
Depreciated: old progress bar to be replaced by Progressbar environment
"""
def progressbar(index, total, length=None, title='Please wait', prefix='', end=False):
    import shutil
    if length is None:
        length = shutil.get_terminal_size()[0]-79
    percent_done = round((index+1)/total*100, 1)
    done = round(percent_done/(100/length))
    togo = length-done

    done_str = '>' * int(done) # '�'
    togo_str = ' ' * int(togo)

    if end: prefix = ' '*60

    print('| %-11s %s%s %3.0f%% %s' % (title, done_str, togo_str, percent_done, prefix), end='\r')
    sys.stdout.flush()

    #if end: print('')


"""
print()
    Why?
        Overwrite print() with more features, such as:
        - flush, i.e., no buffer between prints
        - first argument may determine icon and color
    Usage:
        print(category, ...)
    Example #1:
        print('P','thing')
        # Result: | thing
    Example #2:
        print('some','thing')
        # Result: some thing
"""
print_categories = {
    'info': ['i',''],
    'Err':  ['!','red'],
    'Warn': ['!','magenta'],
    'Msg':  ['i','cyan'],
    'Read': ['>','yellow'],
    'Out':  ['<','green'],
    'Proc': ['|',''],
    'Done': ['=','']}


_orig_print = print
def print(*args, **kwargs):

    prefix = ''
    fmt = ''
    # Check if first argument looks like 'info' or 'Read' or any print category
    if len(args)>1 and isinstance(args[0], str) and args[0].strip()!='':
        # in some cases only change the color
        add_prefix = True
        if '_only_color' in args[0]:
            c = args[0][0:args[0].index('_only_color')]
            add_prefix = False
        else:
            c = args[0]
        # Any category starts with it?
        for category in print_categories.keys():
            if category.startswith(c):
                # Change prefix icon
                if add_prefix:
                    prefix = print_categories[category][0]+' '
                # Change color
                fmt    = print_categories[category][1]
                args = args[1:]
                break

    s = cprintf(prefix + ' '.join(map(str, args)), fmt)
    _orig_print(s, flush=True, **kwargs)

# Notes maybe useful for later improvements of print():
    #def print(*args, **kwargs):
    #    _orig_print(*args, flush=True, **kwargs)

    #print_indent_step = 2
    #print_verbose = 2
    #print_indent = 0
"""
print()
    verbose:
        1 = sometimes
        2 = more often
        3 = even more often
        ...
    Usage:
        print(verbose, level, category, string, other)
    Example:
        print(1, 3,'info','My info text')
        # Result: i       My info text
"""
    # Notes: evaluating vic by first arguments: print(1,2,'p',...)
    # Check for format: verbose, category, level, string
    #if len(args)>1 and isinstance(args[0], int):
    #    verbose  = args[0]
    #    args = args[1:]
    #    if len(args)>1 and isinstance(args[0], int):
    #        indent = ' '*(print_level_step * args[0])
    #        args = args[1:]
    #        if len(args)>1 and isinstance(args[0], str) and args[0] in print_categories:
    #            category = args[0]
    #            prefix = print_categories[category][0]+' '
    #            fmt = print_categories[category][1]
    #            args = args[1:]

    # Notes: evaluating vic by first argument: print('12p',...)
    #verbose = print_verbose
    #prefix = ''
    #indent = print_indent * print_indent_step
    #fmt = ''
    #if len(args)>1 and isinstance(args[0], str) and len(args[0])==3:
    #    # Likely that it means vic (verbose, indent, category)
    #    try:
    #        verbose  = int(args[0][0])
    #        indent = print_indent + print_indent_step * int(args[0][1])
    #        c  = args[0][2]
    #        args = args[1:]
    #        for category in print_categories.keys():
    #            if category.startswith(c):
    #                prefix = print_categories[category][0]+' '
    #                fmt    = print_categories[category][1]
    #    except:
    #        pass
    #if verbose <= print_verbose:
    #s = cprintf(prefix + ' '*int(indent) + str(args[0]), fmt)
