"""
CoRNy Config
    Deal with Configuration files
"""

import os
import configparser
from corny.basics import print, Progress, safe_cast, xsplit, str2dt
from corny.io import format_path, make_dir

class Config():
    def __init__(self, file=None, chdir=True,
                format_path = [('input','path'),('output','out_path')],
                make_dir = [('output','out_path')]):
        if file is None: file = 'config.cfg'
        self.file = file
        self._chdir = chdir
        self._format_path = format_path
        self._make_dir = make_dir
        self.config = None

    def read(self):
        self.config = read_config(self.file)

        if self.config:
            if self._chdir:
                config_dir, config_file_name = os.path.split(self.file)
                os.chdir(os.path.normpath(config_dir))

            if self._make_dir:
                for sec, key in self._make_dir:
                    make_dir(self.get(sec, key))

            if self._format_path:
                for sec, key in self._format_path:
                    self.config[sec][key] = format_path(self.get(sec, key))

        return(self)

    def check(self, sec, key, value=None):
        return(cc(self.config, sec, key, value))

    def get(self, sec, key, alt=None, dtype=None, range=None, types='str'):
        return(gc(self.config, sec, key, alt, dtype, range, types))

    def str(   self,    sec, key, alt=None):
        return(self.get(sec, key, alt, dtype=str))
    def int(   self,    sec, key, alt=None):
        return(self.get(sec, key, alt, dtype=int))
    def float( self,    sec, key, alt=None):
        return(self.get(sec, key, alt, dtype=float))
    def bool(  self,    sec, key, alt=None):
        return(self.get(sec, key, alt, dtype=bool))
    def list(  self,    sec, key, alt=None, range=None, types='str'):
        return(self.get(sec, key, alt, dtype=list, range=range, types=types))
    def date(  self,    sec, key, alt=None):
        return(self.get(sec, key, alt, dtype='date'))
    def multi( self,    sec, key, alt=None):
        return(self.get(sec, key, alt, dtype='multi'))


def read_config(file, chdir=False):
    # Config
    config = configparser.ConfigParser()
    if file is None: file = 'config.cfg'

    with Progress('Loading config file: %s' % file, fmt='R') as Pr:
        if not os.path.isfile(file):
            Pr.fail()
            print('Err','Config file required! Either specify it as an argument or provide ./config.cfg')
            return()
        else:
            temp = config.read(file)

    return(config)


def check_config(config, section, key, value=None):
    # return False if item does not exist or is empty
    r = False
    if section in config:
        if key in config[section]:
            if value is None:
                if config[section][key].strip() != '':
                    r = True
            elif config[section][key].strip() == value:
                r = True
    return(r)

cc = check_config

def get_config(config, section, key, alt=None, dtype=None, range=None, types=str):
    # return None if item does not exist or is empty
    r = alt
    if section in config:
        if key in config[section]:
            if config[section][key].strip() != '':
                r = config[section][key].strip()

    if not r is None:
        if not dtype is None:
            if dtype=='multi':
                r = [item for item in r.split('\n')]
            elif dtype=='date':
                r = str2dt(r)
            elif dtype==list:
                r = xsplit(r, range=range, type=types)
            elif dtype==bool:
                r = yesno2bool(r)
            else:
                r = safe_cast(r, dtype)

    if r is None and dtype==list:
        r = []
    return(r)

gc = get_config

def yesno2bool(s):
    if isinstance(s, str):
        s = s.lower().strip()
        if s == 'yes' or s == 'y' or s=='true' or s=='1':
            return(True)
        else:
            return(False)
    else:
        return(False)
