"""
CoRNy input/ouput
    Toolbox for file operations like download, read, parse, and write
"""
from corny.basics import print, Progress, Progressbar, xsplit, csplit
import pandas
from glob import glob1

def make_dir(dir):
    if not os.path.exists(dir):
        with Progress('Creating folder %s' % dir):
            os.makedirs(dir)

def format_path(dir):
    if os.path.isdir(dir):
        # add trailing slash
        return(os.path.join(dir, ''))
    else:
        return(dir)


def read_files_as_str(path, files, columns=[], skip=0,
                      data_filter=['startswith_digit'],
                      header_filter=['startswith_RecordNum','startswith_GpsUTC']):
    """
    Takes a list of files, calls read() for each, and append to string
    """

    if isinstance(columns, str):
        if columns.strip() != '':
            columns = xsplit(columns)

    archive = None
    if path.endswith('.zip'):
        try:
            archive = zipfile.ZipFile(path, 'r')
        except Exception as e:
            print('Err','Cannot read zip archive: '+str(e))

    datastr = ''
    with Progressbar(len(files), 'Reading') as Pb:
        for filename in files:
            Pb.update(filename)
            if (columns is None or columns == []) and datastr=='':
                columns = read_header(filename, archive, path=path, filters=header_filter)
            datastr += read_file_as_str(filename, archive, path=path, skip=skip, filters=data_filter)

    if archive: archive.close()
    return(datastr, columns)


def read_files(pattern, index_col=1, skiprows=14):
    from glob import glob
    D = None
    for f in glob(pattern):
        Dx = pandas.read_csv(f, index_col=index_col, parse_dates=True, infer_datetime_format=True,
                             skipinitialspace=True, sep=',', skiprows=skiprows, encoding="cp850",
                       names=['RecordNum','t','#ClkTicks','PTB110_mb','P4_mb','P1_mb','Vbat','T1_C','RH1',
                              'N1Cts','N2Cts','N1ETsec','N2ETsec','N1T(C)','N1RH','N2T(C)','N2RH','T_CS215','RH_CS215',
                              'GpsUTC','LatDec','LongDec','Alt','Qual','NumSats','HDOP',
                              'Speed_kmh','COG','SpeedQuality','strDate'])
        if D is None:
            D = Dx.copy()
        else:
            D = pandas.concat([D, Dx])
    return(D)
    #D.head(10)

#######################################
#   Files in/out                      #
#######################################


def file_save_msg(file, name='File', end=''):
    size = os.stat(file).st_size
    if   size < 1024:    sizestr = "%.0f Bytes" % (size)
    elif size < 1024**2: sizestr = "%.0f KB"    % (size/1024)
    else:                sizestr = "%.1f MB"    % (size/1024/1024)
    print('Out','%s saved to %s (%s)' % (name, file, sizestr), end=end)


def download(archive_file='', source='FTP', server=None, user=None, pswd=None,
             ftp_prefix='', ftp_suffix='', sd_path='', sd_prefix='', sd_suffix='',
             update_all=False):

    # if a folder was given
    if not archive_file.endswith('zip'):
        archive_file = os.path.join(archive_file,'') + 'archive.zip'

    # Check what has already been downloaded
    dirname = os.path.dirname(archive_file)
    if dirname != '' and not os.path.isdir(dirname):
        os.makedirs(dirname)
    archive = zipfile.ZipFile(archive_file, 'a', zipfile.ZIP_DEFLATED)
    archive_list = archive.namelist()
    archive_size = os.stat(archive_file).st_size /1024/1024
    print("i","%5s files were already downloaded (%.2f MB)." % (len(archive_list), archive_size))

    if source=='FTP':
        # Select new files from FTP
        import ftplib
        try:
            with Progress('FTP connection') as Pr:
                remote = ftplib.FTP(server)
                Pr.update('OK, login')
                remote.login(user, pswd)

                Pr.update('OK, looking for: %s*' % ftp_prefix)
                if '/' in ftp_prefix:
                    folder = ftp_prefix[0:ftp_prefix.rindex('/')]
                    ftp_prefix = ftp_prefix[ftp_prefix.rindex('/')+1:len(ftp_prefix)]
                    remote.cwd(folder)
                    Pr.update('folder: '+folder)

                remote_list = list(filter(lambda item: item.startswith(ftp_prefix), remote.nlst()))

            print("i","%5s files found on the FTP server." % len(remote_list))

        except ftplib.all_errors as e:
            print('Err','FTP connection error: %s' % e)
            exit()

    elif source=='SD':
        # Select new files from SD-Card
        #remote_list = list(filter(lambda item: item.endswith(suffix=''), os.listdir(sd_path)))
        remote_list = [os.path.basename(x) for x in glob(sd_path+'/'+sd_prefix+'*'+sd_suffix)]
        print("i","%5s files found on the SD backup." % len(remote_list))

    elif source=='DMP':
        # Select new files from DMP
        remote_list = [os.path.basename(x) for x in glob(sd_path+'/'+sd_prefix+'*'+sd_suffix)]
        print("i","%5s files found on the SD backup." % len(remote_list))


    else:
        return

    if update_all:
        update_list = remote_list
    else:
        update_list = list(filter(lambda item: not item in archive_list, remote_list))
    print("i %5s new files selected for download." % len(update_list))

    if len(update_list) == 0: return

    # Download files and add to Zip
    from io import BytesIO
    with Progressbar(len(update_list),'Archiving') as Pb:
        for filename in update_list:
            #print("|  Archiving files %5s/%s: %s" % (i, len(update_list), filename))
            Pb.update(filename)

            if source=='FTP':
                memory = BytesIO()
                remote.retrbinary("RETR " + filename, memory.write) # FTP->memory
                archive.writestr(filename, memory.getvalue())    # memory->archive

            elif source=='SD':
                archive.write(os.path.join(sd_path,'') + filename, filename)

    archive.close()
    if source=='FTP': remote.close()
    file_save_msg(archive_file, "Archive", end="\n")


def select_files(path='', prefix='', suffix=''):
    """
    select_files()
        Why?
            Select files from folder or zip archive which match a pattern
        Note:
            For future gui ouput, add:
            if 'gui' in globals():
                gui.Popup("Input path is neither a folder nor a zip archive.", title='Read ERROR', non_blocking=True)
    """

    _is_archive = path.endswith('.zip')
    if not isinstance(prefix, str): prefix = ''
    if not isinstance(suffix, str): suffix = ''

    pattern = prefix+'*'+suffix
    print('R','Reading data from %s%s%s' % (path, '/' if _is_archive else '', pattern))
    files = []

    if _is_archive:
        try:
            archive = zipfile.ZipFile(path, 'r')
            files = [f for f in archive.namelist() if f.endswith(suffix) and f.startswith(prefix)]
            archive.close()
        except Exception as e:
            print('Err','Cannot read zip archive:', str(e))

    elif os.path.isdir(path):
        files = glob1(path, pattern)
    else:
        print('Err','Input path is neither a folder nor a zip archive.')
        return(None)

    if len(files) == 0:
        print('Err','No files found that match the pattern.')
        return()
    else:
        print('i','Found %d matching files.' % len(files))
        return(files)

import re
re_dataselect = re.compile(r'//DataSelect\s*=\s*(\w+)')
re_columns = re.compile(r'/+(RecordNum.+)')
re_columns2 = re.compile(r'/+(GpsUTC.+)')
re_crnsdata = re.compile(r'^\d+')

def read_file_as_str(filename, archive=None, path='', skip=0, filters=None):

    header_skip = 0
    datastr = ''
    with archive.open(filename,'r') if archive else open(path+filename, encoding="cp850") as file:
        for _ in range(skip):
            next(file)

        for line in file:
            if isinstance(line, bytes):
                line = line.decode(errors='ignore')
            # Check for valid data
            #if re_crnsdata.search(line):

            if check_line(line, filters): #'startswith_digit'):
                datastr += line

    return(datastr+"\n")


import re
import os
re_crnsdata = re.compile(r'^\d+\,')
re_strangecoord = re.compile(r'\d,[EW],')
re_columns = re.compile(r'/+(RecordNum.+)')
re_columns2 = re.compile(r'/+(GpsUTC.+)')

def check_line(line, filters=[]):

    if isinstance(filters,str):
        filters = xsplit(filters)

    for filter in filters:
        if filter=='startswith_digit':
            return(line[0].isdigit())
        elif filter=='strange_coords':
            return([line.count(',', 0, m.start())+1 for m in re_strangecoord.finditer(line)])
        elif filter=='startswith_RecordNum':
            m = re_columns.search(line)
            if m:
                return(m.group(1))
            else:
                return(False)
        elif filter=='startswith_GpsUTC':
            m = re_columns2.search(line)
            if m:
                return(m.group(1))
            else:
                return(False)

    return(True)


def read_header(filename, archive=None, path='', filters=None):

    strange_coords = []
    columns = []

    with archive.open(filename,'r') if archive else open(path+filename) as file:

        for line in file:
            if isinstance(line, bytes):
                line = line.decode(errors='ignore')

            if check_line(line, 'startswith_digit'):
                # this is the first valid data line. Does it contain strange EW coordinates?
                strange_coords = [line.count(',', 0, m.start())+1 for m in re_strangecoord.finditer(line)]
                break

            elif not filters is None:
                for filter in filters:
                    r = check_line(line, filter)
                    if r:
                        columns.append(r)
                        continue

    if len(columns) <= 0:
        print('E','Input file contains no headers (e.g., //RecordNum...). Try to define input_columns in the config file.')
        return(None)
    else:
        columns = csplit(chomp(','.join(columns)))

        if strange_coords:
            print("\n  "+'Found strange coordinate format and missing columns, trying to fix it...', end='')
            new_columns = ['NS','EW'] # Assuming that northing is first, easting is second.
            for pos in strange_coords:
                columns.insert(pos, new_columns.pop(0))

        return(columns)


def read_file(filename, archive=None, tz='UTC'):
    header_skip = 0
    datastr = ''
    with archive.open(filename,'r') if archive else open(filename, encoding="cp850") as file:
        for line in file:
            if isinstance(line, bytes):
                line = line.decode(errors='ignore')
            if re_crnsdata.search(line):
                datastr += line
    return(datastr+"\n")

import io
import os
import zipfile

def read_and_convert(path, prefix='', suffix='', data_columns=[], progress='file'):

    archive = None
    files = []

    if path.endswith('.zip'):
        archive = zipfile.ZipFile(path, 'r')
        files = [x for x in archive.namelist() if x.endswith(suffix) and x.startswith(prefix)]

    elif os.path.isdir(path):
        pattern = path + prefix +'*'+ suffix
        files = glob(pattern)
        if len(files)==0:
            print("Err: Folder is empty: " + pattern)
            return(None)
    else:
        print("Err","Input is neither a file, folder, nor a zip archive.")
        return(None)

    datastr = ''
    i = 0
    last_progress = 0 # in %
    i_max = len(files)

    for filename in files:
        i += 1

        if progress == 'percent':
            this_progress = (i+0.01)/len(files)//0.1*10 # in %, offset 0.01 because x/x//0.1 is 9
            if this_progress > last_progress:
                print("%.0f%% " % this_progress, end='')
                last_progress = this_progress

        elif progress == 'file':
            print('%5d/%d: %s' % (i, i_max, filename))

        if datastr=='':
            #if data_columns == []:
            #    data_columns = read_header(filename, archive)
            datastr = read_file(filename, archive)
        else:
            datastr += read_file(filename, archive) #data.append() #, verify_integrity=True)

    if not archive is None:
        archive.close()

    #DEBUG output raw data file
    #with open('temp.txt','a') as f:
    #    f.write(datastr)

    data = make_pandas(datastr, data_columns)
    if data is None:
        print('Err','Cannot interprete the data. Check the column header, is this correct? '+ ','.join(data_columns))
        return()
    else:
        return(data)

def make_pandas_old(datastr, file=False, columns=None, header_skip=0, index=1):
    if file:
        f = datastr
    else:
        f = io.StringIO(datastr)

    try:
        data = pandas.read_csv(f, index_col=index, parse_dates=True, infer_datetime_format=True, skipinitialspace=True,
                               sep=',', skiprows=12, encoding="cp850", error_bad_lines=False, warn_bad_lines=False, names=columns, dtype=object)
    except Exception as e:
        print('E','interpreting data format: '+str(e))
        return(None)

    data = data.apply(pandas.to_numeric, errors='coerce')
    return(data)


def make_pandas(datastr, columns=None, header_skip=0, index=1, sep=',', timestamp=False):
    """
    make_pandas()
        Why?
            Convert datastr, e.g. from a file, to a clean pandas time series
        Note:
            Depreciated: no longer convert columns directly on input as it is error-prone:
            parse_dates=True, infer_datetime_format=True, index_col=index, dtype = dtypes
            dtypes = { c : float for c in columns }; dtypes[columns[index]] = str
    """

    with Progress('  Parsing data string as Pandas object   ') as Pr:
        data = pandas.read_csv(io.StringIO(datastr), encoding='cp850', skipinitialspace=True,
                                sep=sep,  skiprows=header_skip, names=columns,
                                error_bad_lines=False, warn_bad_lines=False,
                                dtype=object)
        Pr.update('(%5d lines x %2d columns)' % (len(data),len(data.columns)))

    if isinstance(index,str):
        il = xsplit(index, type=int)
        if len(il) == 1:
            index = il[0]
        else:
            index = list(il)

    with Progress('  Converting index column %2s to datetime ' % str(index)) as Pr:
        data = data.loc[data.index.dropna()]
        if isinstance(index, list):
            if len(index) == 2:
                index_column = '_time'
                #TODO: check if index out of bounds
                Pr.update('merging %s and %s ' % (index[0], index[1]))
                data[index_column] = pandas.to_datetime(data.iloc[:,index[0]] +' '+ data.iloc[:,index[1]], errors='coerce')
            else:
                pass #print(index)
        else:
            index_column = columns[index]
            if timestamp:
                data[index_column] = pandas.to_datetime(data[columns[index]], errors='coerce', unit='s')
            else:
                data[index_column] = pandas.to_datetime(data[columns[index]], errors='coerce')

        ## Drop nan in index
        data = data.dropna(subset=[index_column])
        ## Set the DateTime column as index
        data.set_index(index_column, inplace=True)

    with Progress('  Sort data by date and drop duplicates  '):
        data = data.sort_index().drop_duplicates().tz_localize('UTC')

    with Progress('  Convert columns to numeric, drop cheese'):
        # to numeric
        data = data.apply(pandas.to_numeric, errors='coerce')

        ## Find out which column names are null
        bad_columns = []
        for c in range(len(columns)):
            if pandas.isnull(columns[c]):
                bad_columns.append(c)
        ## Reassign data using only non-bad columns
        data = data.iloc[:, [i for i,n in enumerate(data.columns) if i not in bad_columns]]

        Pr.update('(%5d lines x %2d columns)' % (len(data),len(data.columns)))

    return(data)

def check_pandas(data):
    if data is None:
        print('Err','Invalid data object, maybe check headers and period.')
    elif len(data)==0:
        print('Err','No remaining records in dataset, maybe check headers and period.')
    else:
        return(True)
    return(False)


def read_corny(file):
    data = pandas.read_csv(file, index_col=0, parse_dates=True, infer_datetime_format=True, skipinitialspace=True,
                           sep=',', encoding="cp850", error_bad_lines=False, warn_bad_lines=False) #, dtype=object)
    print("%i lines, %i columns: %s" % (len(data), len(data.columns), ', '.join(data.columns)))
    return(data)

read_cornish = read_corny

def read_geojson(file, columns=['label','short','type','lat','lon']):
    import geojson
    with open(file) as f:
        gj = geojson.load(f)

    # Run through geojson and fill a Sensor DataFrame
    D = pandas.DataFrame([], columns=columns)
    for i in range(len(gj['features'])):
        Dx = pandas.DataFrame([[
            gj['features'][i]["properties"]['Id'],
            gj['features'][i]["properties"]['short'],
            gj['features'][i]["properties"]['Inst'],
            gj['features'][i]["geometry"]["coordinates"][1],
            gj['features'][i]["geometry"]["coordinates"][0]]],
            columns=columns)
        D = D.append(Dx)
    D = D.sort_values('label')
    D['color'] = 'black'
    D = D.set_index(np.arange(1,len(D)+1))
    return(D)



def report_size(file, verb='Saved', unit='MB'):
    import os
    if unit=='KB':
        factor = 1/1024
    elif unit=='MB':
        factor = 1/1024/1024
    else:
        factor = 1
        unit = 'unknown'
    size = os.stat(file).st_size * factor
    print("%s %s (%.1f %s)." % (verb, file, size, unit))
