#!/usr/bin/env python
VERSION = '0.5.9'

from corny import *
from corny.exdata import *
from corny.exdata.raster import luse_code2str, get_from_raster
import corny.exdata as exdata
from corny.exdata import footprint

def main(configfile=None):
    print("Welcome to instant-PASDy %s!" % VERSION)

    # Find script path
    pasdy_path = os.path.dirname(os.path.abspath(__file__))

    # Read config
    C = Config(configfile).read()
    config = C.config
    if not config:
        exit()

    # Update data (FTP, SD, ...)
    if 'update' in config:
        data_source = csplit(config['update']['data_source'])
        if 'SD' in data_source or 'FTP' in data_source:
            print('Update')
            for source in data_source:
                print('Proc','Updating data from ' + source + ' into ' + config['input']['path'])
                download(
                    archive_file = config['input']['path'],        source     = source,
                    server       = config['update']['FTP_server'], user       = config['update']['FTP_user'],
                    pswd         = config['update']['FTP_pswd'],   ftp_prefix = config['update']['FTP_prefix'],
                    ftp_suffix   = config['update']['FTP_suffix'], sd_path    = config['update']['SD_path'],
                    sd_prefix    = config['update']['SD_prefix'],  sd_suffix  = config['update']['SD_suffix'],
                    update_all = yesno2bool(gc(config,'update','update_all','no')))

    coordinates = xsplit(gc(config,'info','coords'), range=3, type=float)

    ####################################################
    ######  Input  #####################################
    ####################################################

    print('Input')

    files = select_files(path  =C.get('input','path'),
                         prefix=C.get('input','prefix'),
                         suffix=C.get('input','suffix'))

    datastr, columns = read_files_as_str(path = C.get('input','path'),
                         files=files, columns=C.list('input','input_columns'),
                         skip=C.int('input','skip_lines', alt=0),
                         data_filter=['startswith_digit'], header_filter=['startswith_RecordNum','startswith_GpsUTC'])

    data = make_pandas(datastr, columns, index=C.get('input','index_column_number',alt=1),
            sep=C.get('input','sep',alt=','), timestamp=C.bool('input','timestamp',alt=''))

    if not check_pandas(data): exit()

    if C.bool('input','separate_gps'):
        print('Input GPS')

        gps_files = select_files(path=C.get('input','path'),
                             prefix=C.get('input','gps_prefix'),
                             suffix=C.get('input','gps_suffix'))

        gps_datastr, gps_columns = read_files_as_str(path = C.get('input','path'),
                             files=gps_files, columns=C.list('input','gps_input_columns'),
                             skip=C.int('input','gps_skip_lines', alt=0),
                             data_filter=['startswith_digit'])

        gps_data = make_pandas(gps_datastr, gps_columns, index=C.get('input','gps_index_column_number',alt=1),
                sep=C.get('input','sep',alt=','), timestamp=C.bool('input','gps_timestamp',alt=''))

        if not check_pandas(gps_data): exit()

        data = merge_datasets(data, gps_data, 'input', 'GPS')

    # Cut period
    data = cut_period_to(data, C.get('clean','start'), C.get('clean','end'))
    if not check_pandas(data): exit()

    print('i','Data period from', data.index.min().strftime('%Y-%m-%d %H:%M:%S'),
                            'to', data.index.max().strftime('%Y-%m-%d %H:%M:%S'))

    ####################################################
    ######  Basic cleaning  ############################
    ####################################################

    print('Quality check')

    bbox = C.list('clean','bbox', types='float')

    # Set new data columns as copy of old ones
    #data['Lat']  = new_column(data, copy_from=C.get('input','latitude_column'),  dispense=True)
    #data['Lon']  = new_column(data, copy_from=C.get('input','longitude_column'), dispense=True)
    #data['Alt']  = new_column(data, copy_from=C.get('input','altitude_column'),  dispense=True)
    data['Pres'] = new_column(data, copy_from=C.get('input','pressure_column'))
    data['rHum'] = new_column(data, copy_from=C.get('input','humidity_column'))
    data['Temp'] = new_column(data, copy_from=C.get('input','temperature_column'))
    #ah =  gc(config, 'correction', 'new_humidity_column')
    #inc = gc(config, 'correction', 'new_incoming_column')

    # Thermal neutrons
    th = '_thermal'
    # Cleaned data
    clean = '_clean'
    # Corrected data
    pih = '_pih'
    # Suffix for vairable errors
    err = '_err' # followed by _sym, _low, _upp
    # Standard deviation
    std = '_std'
    # Smooth
    mov = '_mov'
    # Biomass
    bio = '_bio'
    # Urban
    urb = '_urb'
    # Road
    road = '_road'

    data['resolution'] = new_column(data, copy_from=C.get('input', 'resolution_column')) #, alt='N1ET_sec'))
    if data['resolution'].isnull().all():
        data['resolution'] = infer_tres(data, default=C.int('input', 'resolution_default'), in_sec=True)
    # TODO: make diagnistic plot for resolutio versus resolution column

    ### Clean neutrons

    neutron_columns = C.list('input','neutron_columns')
    for tube in neutron_columns:
        # Stochastic Uncertainty
        data[tube+'_err'] = neutron_uncertainty(data[tube], data['resolution'], units=C.get('clean','neutrons_unit'))

        # Convert to cph
        data[tube+'_cph'] = convert_units(data[tube], units_from=C.get('clean','neutrons_unit'), units_to='cph', resolution=data['resolution'])
        data[tube+'_cph_err'] = convert_units(data[tube+'_err'], units_from=C.get('clean','neutrons_unit'), units_to='cph', resolution=data['resolution'])

        # Filter for outlier
        neutron_range_per_tube = C.list('clean','neutron_range_per_tube', range=2, types=float)
        data[tube+'_c'] = find_outliers(data[tube+'_cph'], lower=neutron_range_per_tube[0], upper=neutron_range_per_tube[1],
                                        n_sigma=C.float('clean','neutron_n_sigma'), sigma=data[tube+'_cph_err'])
        data[tube+'_c_err'] = data[tube+'_cph_err']

    data['N_cph']   = sum_neutrons(data[[tube+'_cph' for tube in neutron_columns]])
    data['N_c']     = sum_neutrons(data[[tube+'_c' for tube in neutron_columns]])
    data['N_c_err'] = sum_neutrons_err(data[[tube+'_c_err' for tube in neutron_columns]])

    # TODO: do that for thermal too

    # Report
    report_N(data, 'N_c')

    # Thermal neutrons
    if config['input']['thermal_neutron_columns']:

        # sum up thermal neutrons
        data[N+th] = 0
        for c in xsplit(config['input']['thermal_neutron_columns']):
            data[N+th] += data[c]

        data[N+th+err] = N_err(data[N+th])

        # convert to units of cph
        data[N+th]     = Unit_convert(data[N+th], data[tres], 'cph')
        data[N+th+err] = Unit_convert(data[N+th+err], data[tres], 'cph')

        Nth = N+th
    else:
        Nth = None

    ####################################################
    ######  Clean/filter  ##############################
    ####################################################

    print('| Cleaning... ')

    has_coords = (lat and lon) or not np.isnan(coordinates[0:2]).any()

    # Clean altitude
    if not alt:
        if has_coords:
            data[alt] = coordinates[2]
        else:
            print('i No coordinates specified, assuming altitude = 0 m.')
            data[alt] = 0.0

    # Clean Latlon
    if lat != '' and lat in data.columns:
        #JJ Strangely formatted coordinates occur for some Rovers (Jülich, Melbourne)

        if 'EW' in data.columns or 'NS' in data.columns:
            # Remove cardinal direction if it is attached to the coordinate (has been necessary in some configurations)
            data[lat] = pd.to_numeric(data[lat].astype(str).replace({'N':''}, regex=True), errors='coerce') # replace N if existent in column and convert to numeric
            data[lon] = pd.to_numeric(data[lon].astype(str).replace({'E':''}, regex=True), errors='coerce') # replace E if existent in column and convert to numeric
            # Convert to readable coordinate
            data[lat] = deg100min2dec( data[[lat, 'NS']] )
            data[lon] = deg100min2dec( data[[lon, 'EW']] )

        data.loc[data[lat] == 0.0, lat] = np.nan
        data.loc[data[lon] == 0.0, lon] = np.nan

        data['lon'] = data[lon]
        data['lat'] = data[lat]

        if 'interpolate_coords' in config['clean']:
            if yesno2bool(config['clean']['interpolate_coords']):
                data['lon'] = data['lon'].interpolate()
                data['lat'] = data['lat'].interpolate()

        data = data.dropna(subset=['lat'])

        if len(bbox)>0:
            data = data.query("lat > %s and lat < %s" % (bbox[2], bbox[3]))
            data = data.query("lon > %s and lon < %s" % (bbox[0], bbox[1]))
            if len(data) == 0:
                print(cprintf('! You seem to have erased the whole dataset using an unfortunate bbox.', 'red'))
                return()
        else:
        # Derive bbox from data if not set in config
            bbox = [data[lon].min(), data[lon].max(), data[lat].min(), data[lat].max()]
            print('i   Derived bounding box:', bbox)

        try:
            data['ll'] = data.apply(lambda row: tuple((row['lat'],row['lon'])), axis=1)
        except ValueError:
            raise ValueError('NO GPS Data Available, processing stopped')
        #D = D.dropna(subset=['lat'])
        # Lat lon to UTM
        utmy, utmx = latlon2utm(data.lat, data.lon)
        data['utmy'] = utmy
        data['utmx'] = utmx


        # if no geographic information available we stop processing here
        if data[lat].isna().sum()==len(data) or data[lon].isna().sum()==len(data):
            print(cprintf('! Cannot find any GPS data )\':', 'red'))
            raise ValueError('No GPS data available, processing stopped.')

        # replace altitude column with data from DEM
        if 'altitude_raster' in config['clean'] and config['clean']['altitude_raster']:
            data = get_from_raster(data, alt, config['clean']['altitude_raster'])
            print('i Derived altitude: %.0f +/- %.0f m' % (data[alt].mean(), data[alt].std()))

    # Get upper and lower bounds from config
    (Na,Nb)  = xsplit(config['clean']['neutron_range'], range=2, type=float)
    (pa,pb)  = xsplit(config['clean']['pressure_range'], range=2, type=float)
    (rha,rhb)= xsplit(config['clean']['humidity_range'], range=2, type=float)
    (Ta,Tb)  = xsplit(config['clean']['temperature_range'], range=2, type=float)
    if 'timeres_range' in config['clean']:
        (tra,trb)= xsplit(config['clean']['timeres_range'], range=2, type=float)

    # Define new column N+clean
    data[N+clean] = data[N]
    data[N+clean+err] = data[N+err]
    Ncleanstr = N+clean

    if Nth:
        data[N+th+clean] = data[N+th]
        data[N+th+clean+err] = data[N+th+err]

    # Drop lines option
    if config['clean']['invalid_data'] == 'drop lines':
        len_before = len(data)
        data = data.query("@Ncleanstr > %s and @Ncleanstr < %s" % (Na, Nb))
        data = data.query("@p > %s and @p < %s" % (pa, pb))
        data = data.query("@rh> %s and @rh< %s" % (rha,rhb))
        data = data.query("@T > %s and @T < %s" % (Ta, Tb))
        if 'timeres_range' in config['clean']:
            data = data.query("@tres > %s and @tres < %s" % (tra, trb))
        print("dropped %s invalid lines." % (len_before - len(data)))
    # Fill-na option
    else:
        nan_before = data.isnull().sum().sum()
        local_storage_name = os.path.join(config['output']['out_path'],'') + config['output']['out_basename']

        # Clean range: neutrons
        data[Ncleanstr] = find_outlier(data, Ncleanstr, lower=Na, upper=Nb,
                                       n_sigma = gc(config,'clean','neutron_n_sigma', dtype=float),
                                       sigma = Ncleanstr+err)
        if Nth:
            data[N+th+clean] = find_outlier(data, N+th+clean, lower=Na, upper=Nb,
                                           n_sigma = gc(config,'clean','neutron_n_sigma', dtype=float),
                                           sigma = N+th+clean+err)
        # Clean range: time resolution
        if cc(config, 'clean', 'timeres_range'):
            data[Ncleanstr] = find_outlier(data, tres, Ncleanstr, lower=tra, upper=trb)

        # Clean range: air humidity
        data[rh] = find_outlier(data, rh, lower=rha, upper=rhb)
        data[rh] = fill_gaps(data, rh, var='humidity', fill=gc(config, 'clean', 'missing_humidity'),
                            n_stations=4, lon=one_of(lon, coordinates[1]), lat=one_of(lat, coordinates[0]),
                            local_storage_name=local_storage_name)

        # Clean range: air temperature
        data[T] = find_outlier(data, T, lower=Ta, upper=Tb)
        data[T] = fill_gaps(data, T, var='temperature', fill=gc(config, 'clean', 'missing_temperature'),
                            n_stations=4, lon=one_of(lon, coordinates[1]), lat=one_of(lat, coordinates[0]),
                            local_storage_name=local_storage_name)

        # Clean range: air pressure
        if cc(config,'clean','pressure_unit','Pa'):
            data[p] /= 100

        data[p] = find_outlier(data, p, lower=pa, upper=pb)
        data[p] = fill_gaps(data, p, var='pressure', fill=gc(config, 'clean', 'missing_pressure'),
                            n_stations=4, lon=one_of(lon, coordinates[1]), lat=one_of(lat, coordinates[0]),
                            alt = alt, T = T, local_storage_name=local_storage_name)

        print("i Replaced %s values by NaN." % (data.isnull().sum().sum() - nan_before))

    if len(data)==0:
        raise ValueError('! No valid data left :\'(')


    # Interpolate over nans
    if yesno2bool(config['clean']['interpolate_neutrons']):

        max_gap_size = gc(config,'clean','interpolate_neutrons_gapsize',dtype=int)

        data[N+clean]     = data[N+clean].interpolate(limit=max_gap_size)
        data[N+clean+err] = data[N+clean+err].interpolate(limit=max_gap_size)
        if Nth:
            data[N+th+clean]     = data[N+th+clean].interpolate(limit=max_gap_size)
            data[N+th+clean+err] = data[N+th+clean+err].interpolate(limit=max_gap_size)

    report_N(data, N+clean)

    ####################################################
    ######  Soil #######################################
    ####################################################

    print('> Sampling land data... ', end='')
    #depending on Option for soil data retrieval we add soil data
    #print('Retrieve soil properties from constant values')
    #add static data -- this is the default and also the basis
    data['bd']  = float(config['conversion']['bulk_density'])
    data['org'] = float(config['conversion']['soil_org_carbon'])
    data['lw']  = float(config['conversion']['lattice_water'])

    # add from raster datasets
    if 'soil_data_source' in config['conversion'] and config['conversion']['soil_data_source']=='raster':
        print('Retrieve soil properties from static maps')
        if lat in data.columns:
            if config['conversion']['bulk_density_raster']:
                data = get_from_raster(data, 'bd',   config['conversion']['bulk_density_raster'])
            if config['conversion']['clay_content_raster']:
                data = get_from_raster(data, 'clay', config['conversion']['clay_content_raster'])

            if config['conversion']['soil_org_carbon_raster']:
                data = get_from_raster(data, 'org',  config['conversion']['soil_org_carbon_raster'])

        #convert lattice water from clay
        data['lw']  = lw_from_clay(data['clay'], method='Greacen et al. 1981')

    # add from soilgrids
    if cc(config,'conversion','soil_data_source', value='soilgrids'):
        #check whether geographic information exist
        if not has_coords:
            ValueError('No information on coordinates, no retrieval of soil info possible, change input option!')
        else:
            print('Retrieve soil properties from SoilGrids API')
            soil_data=retrieve_soilgrid_data(data.copy(),
                        lon = one_of(lon, coordinates[1]), lat = one_of(lat, coordinates[0]),
                        soilgridlrs=['bdod','clay','soc'],
                        soil_layerdepths=['0-5cm','5-15cm','15-30cm','30-60cm'],
                        raster_res=(250, 250),
                        statistical_metric=['mean'],
                        all_touched=True)

            #add the soil data to the dataset
            data['org']=soil_data['org'] /10 # new units
            data['bd']=soil_data['bd']
            data['clay']=soil_data['clay']

            #convert lattice water from clay
            data['lw']  = lw_from_clay(data['clay'], method='Greacen et al. 1981')

    #convert organic water equivalent from organic matter content
    data['owe'] = owe_from_corg(data['org'], method='Franz et al. 2015')

    #add landuse data
    if lat in data.columns:
        if cc(config,'conversion','land_use_data_source', value='raster'):
            data = get_from_raster(data, 'luse', gc(config,'conversion','land_use_raster'))
        elif cc(config,'conversion','land_use_data_source', value='corine'):
            data['luse'] = exdata.corine.retrieve_corine_data(data,
                            lon = one_of(lon, coordinates[1]),
                            lat = one_of(lat, coordinates[0])
                            ).values.astype(int)
            data['luse_str'] = luse_code2str(data)


    print('bulk dens. (%.2f g/cm3), lattice w. (%.2f g/g), org.we. (%.2f g/g)' % (data.bd.mean(), data.lw.mean(), data.owe.mean()))


    ####################################################
    ######  Incoming  ##################################
    ####################################################

    print(cprintf('> Reading incoming radiation data from ' + config['correction']['NM_path'], 'yellow'), end='')
    M = NM(config=config).get(start=data.index.min().replace(tzinfo=pytz.UTC), end=data.index.max().replace(tzinfo=pytz.UTC))

    data.index = data.index.tz_localize(None)
    data.index = data.index.tz_localize('UTC')
    if isinstance(M.data, pandas.DataFrame):
        M.data.index = M.data.index.tz_localize(None)
        M.data.index = M.data.index.tz_localize('UTC')
        M.data = M.data.reindex(data.index, method='nearest', limit=1, tolerance=timedelta(seconds=M.resolution_min*60/2-1)).interpolate()
        data[inc] = M.data
        data.loc[data[inc] < data[inc].mean()/1.2, inc] = np.nan
        data.loc[data[inc] > data[inc].mean()*1.2, inc] = np.nan
    else:
        data[inc] = float(config['correction']['incoming_ref'])

    ####################################################
    ######  pih  #######################################
    ####################################################

    print("\n| Corrections... ", end='')
    data[ah] = abs_humidity(data[rh], data[T])

    data['correct_h'] = C_humid( data[ah], gc(config, 'correction', 'humidity_ref', dtype=float),
                                           gc(config, 'correction', 'alpha', dtype=float),
                                           method=gc(config, 'correction', 'humidity_method'))

    data['correct_p'] = C_press( data[p],  gc(config, 'correction', 'pressure_ref', dtype=float),
                                           gc(config, 'correction', 'beta', dtype=float),
                                           method=gc(config, 'correction', 'pressure_method'),
                                           inclination = gc(config, 'correction', 'inclination', dtype=float))

    #np.exp((data[p] - float(config['correction']['pressure_ref'])) / float(config['correction']['beta']))
    data['correct_i'] = C_inc( data[inc],  gc(config, 'correction', 'incoming_ref', dtype=float),
                                           gc(config, 'correction', 'gamma', dtype=float),
                                           method=gc(config, 'correction', 'incoming_method'),
                                           Rc = gc(config, 'correction', 'Rc', dtype=float),
                                           Rc_NM = gc(config, 'correction', 'Rc_NM', dtype=float))


    data[N+clean+pih]     = data[N+clean]     * data['correct_p'] * data['correct_i'] * data['correct_h']
    data[N+clean+pih+err] = data[N+clean+err] * data['correct_p'] * data['correct_i'] * data['correct_h']

    if Nth:
        data[N+th+clean+pih]     = data[N+th+clean]     * data['correct_p'] * data['correct_i'] * data['correct_h']
        data[N+th+clean+pih+err] = data[N+th+clean+err] * data['correct_p'] * data['correct_i'] * data['correct_h']
        Nthfinal = N+th+clean+pih

    print("for pressure (%+.0f%%), abs. humidity (%+.0f%%), incoming radiation (%+.0f%%)" % \
        (100*(data['correct_p'].mean()-1), 100*(data['correct_h'].mean()-1), 100*(data['correct_i'].mean()-1)))

    Nfinal = N+clean+pih
    report_N(data, N+clean+pih, correction=['correct_p','correct_i','correct_h'])


    #If no data remain with valid corrected data, we stop processing
    if data[N+clean+pih].isna().sum()==len(data):
        print(cprintf('! No valid neutron data remaining. Are all meteo or GPS data out of range? Or is one of the neutron tubes sick? (then the sum of all tubes would be nan, too).','red'))
        raise ValueError('No valid neutron data remaining.')



    ####################################################
    ######  Urban  #####################################
    ####################################################

    data['correct_urban'] = 1

    if 'luse_str' in data.columns and 'land_use_urban_corr' in config['conversion'] and config['conversion']['land_use_urban_corr'] != '':
        if config['conversion']['land_use_urban_corr'] == 'nan':
            data.loc[data.luse_str=='urban', 'correct_urban'] = np.nan
        else:
            data.loc[data.luse_str=='urban', 'correct_urban'] = float(config['conversion']['land_use_urban_corr'])

        data[Nfinal+urb] = data[Nfinal] * data['correct_urban']
        data[Nfinal+urb+err] = data[Nfinal+err] * data['correct_urban']
        Nfinal += urb
        report_N(data, Nfinal, 'correct_urban')
    else:
        urb = ''

    ####################################################
    ######  Bio  #######################################
    ####################################################
    ## vegetation correct
    ## assume neutron counts were reduced by 10% for every measurement in a forest
    data['correct_bio'] = 1
    data['correct_bio_err'] = 0
    if 'luse_str' in data.columns and 'land_use_forest_corr' in config['conversion'] and config['conversion']['land_use_forest_corr'] != '':
        if config['conversion']['land_use_forest_corr'] == 'nan':
            data.loc[data.luse_str=='forest', 'correct_bio'] = np.nan
        else:
            data.loc[data.luse_str=='forest', 'correct_bio'] = float(config['conversion']['land_use_forest_corr'])

        print("| Correction for biomass (%+.0f%%)." % (100*(data.loc[data.luse_str=='forest', 'correct_bio'].mean()-1)))
        # well do that after road correction
        #data[Nfinal+'_bio'] = data[Nfinal] * data['correct_bio']
        #data[Nfinal+'_bio'+err] = data[Nfinal+err] * data['correct_bio']
        #Nfinal += '_bio'
        #report_N(data, Nfinal, 'correct_bio')

    else:
        bio = ''
        #print('.')


    ####################################################
    ######  Road  ######################################
    ####################################################

    data['correct_road'] = 1
    if config['conversion']['road_method'] != '' and config['conversion']['road_method'] != 'None' and lat in data.columns:

        print('| Correction for road effect...')
        # Load road properties (goes to config?)

        #print('|   Loading road type correction properties...')
        print(cprintf('>   %s' % config['conversion']['road_type_table'], 'yellow'))
        Roads = pd.read_csv(config['conversion']['road_type_table'], sep=r'\s+', skipinitialspace=True, comment='#')

        ####Now we decide whether we get the road_network_Data####
        ####from osmnx or fromroverweb Overpass API####
        if 'road_network_source' in config['conversion'] and config['conversion']['road_network_source']=='osm-nx':
            print('retrieve road network information from osmnx')
            # Load road network
            #print('>   ', end='') # Graph prints a message.
            road_network_file = config['conversion']['road_network_file']

            # Load nearest road types
            road_type_file = os.path.join(config['output']['out_path'],'') + config['output']['out_basename'] +'.roads' #configfile[0:configfile.index('.cfg')]
            print(os.path.join(config['output']['out_path'],'') + config['output']['out_basename'] +'.roads')
            print('|   Loading road network... ', end='')
            if os.path.exists(road_type_file):
                # Read from file
                #print('|   Loading nearest road types...')
                print(cprintf('\n>   %s' % road_type_file, 'yellow'))
                roadsdf = pd.read_csv(road_type_file, index_col=0, parse_dates=True, infer_datetime_format=True)
                data['road'] = roadsdf
            else:
                #download via osmnx, boundary is used from a Polygon, However, bbox is possible as well
                boundary_polygon=footprint.retrieve_footprints(data,footprint_radius=200,lon='LongDec',lat='LatDec',
                        output=None,unify=True)
                R = Graph('roads', boundary_polygon, file=road_network_file, report_save='<   Saved')
                R.get(save=True) # read from file, or download and save

                if len(R.data)<=0:
                    print('i   Road network is empty!')
                else:
                    # Calculate
                    print('|   Searching nearest roads...')
                    # invoke progress bar feature
                    from tqdm import tqdm
                    tqdm.pandas()

                    data['road'] = data.progress_apply(lambda row: R.on_road_type(point=row['ll']), axis=1)

                    #print('    Set priorities...')
                    for i, row in data.iterrows():
                        row = row.copy()
                        if isinstance(row.road, list):
                            if 'primary' in row.road:
                                data.loc[i, 'road'] = 'primary'
                            elif 'track' in row.road:
                                data.loc[i, 'road'] = 'track'
                        #print(D.loc[i, 'road'], end=' ')
                    #print('')

                    # Save backup
                    data['road'].to_csv(road_type_file)
                    file_save_msg(road_type_file, 'Nearest roads', end="\n")

        elif 'road_network_source' in config['conversion'] and config['conversion']['road_network_source']=='osm-overpass':
            print('| Overpass')
            print('retrieve road network information from overpass API')
            data=retrieve_osm_overpass_data(data, lon=lon,lat=lat,crcl_radius=50,
                               no_of_clusters=None,queryfeatures={'way': ['highway']},
                               output_col_name='road')
        # Set road values
        data['road_moisture'] = np.nan
        data['road_width'] = 0.0
        if 'road' in data.columns:
            for i, row in Roads.iterrows():
                row = row.copy()
                data.loc[data.road==row['category'], 'road_moisture'] = row['moisture']
                data.loc[data.road==row['category'], 'road_width']    = row['width']
        else:
            print('|   Applying constant road parameters theta=%s, width=%s' % (config['conversion']['road_moisture'], config['conversion']['road_width']))
            data['road_moisture'] = float(config['conversion']['road_moisture'])
            data['road_width']    = float(config['conversion']['road_width'])
        #pdb.set_trace()
        # correct for biomass to avoid smearing out
        data[Nfinal+'_temp'] = data[Nfinal] * data['correct_bio']
        data = moving_avg(data, Nfinal+'_temp', window=config['clean']['smooth'], err=False, std=False)
        # back to Nfinal, but mavged
        data[Nfinal+'_temp_mov'] /= data['correct_bio']
        # N2SM first order
        if config['conversion']['N0']:
            N0 = config['conversion']['N0']
        else:
            N0 = data.Nfinal.max()*1.1
            print('! N0 is not defined, but we need it for road correction. Guessing %.0f cph.' % N0)
        data['sm_temp'] = N2SM_Desilets(data[Nfinal+'_temp_mov'], N0, bd=data.bd, lw=data.lw, owe=data.owe,
                    a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))
        # C_road
        data['correct_road'] = 1/(1 + 0.42*(1-np.exp(-0.5*data.road_width)) * (1.06-4*data.road_moisture-(0.16 + data.road_moisture)/(0.39 + data['sm_temp'])))
        data.loc[(data['correct_road'] > 1), 'correct_road'] = 1
        #data.loc[data.road_width==0, 'correct_road'] = 1
        data.correct_road = data.correct_road.fillna(1)
        print("i   ...road effect (%+.0f%%)." % (100*(data['correct_road'].mean()-1)))

        # tidy up temp columns
        #data.drop([Nfinal+'_temp', 'sm_temp'], axis=1, inplace=True)

    else:
        road = ''


    # Set N_bio
    data[Nfinal+bio]          = data[Nfinal]         * data['correct_bio']

    if 'land_use_forest_corr_err' in config['conversion']:
        data['correct_bio_err'] = float(config['conversion']['land_use_forest_corr_err'])
    data[Nfinal+bio+err] = errpropag_factor(data, Nfinal, 'correct_bio', err)
    #data[Nfinal+bio+err]      = data[Nfinal+err]     * data['correct_bio']

    Nfinal += bio
    report_N(data, Nfinal, 'correct_bio')

    if config['conversion']['road_method'] != '' and config['conversion']['road_method'] != 'None'  and lat in data.columns:

        # set N_road
        data[Nfinal+road]     = data[Nfinal]     * data['correct_road']

        data['correct_road_err'] = float(config['conversion']['road_correction_err'])
        data[Nfinal+road+err] = errpropag_factor(data, Nfinal, 'correct_road', err)
        #data[Nfinal+road+err] = data[Nfinal+err] / data['correct_road']

        Nfinal += road
        report_N(data, Nfinal, 'correct_road')

    #if config['conversion']['road_moisture'] and float(config['conversion']['road_width']) > 0 and not 'correct_road' in data.columns:
     #           print("| Corrections for road effect... ", end='')
                #if   config['correction']['road_material'] == 'stone':    road_theta = 0.04
                #elif config['correction']['road_material'] == 'concrete': road_theta = 0.07
                #elif config['correction']['road_material'] == 'asphalt':  road_theta = 0.10
      #          road_theta = float(config['conversion']['road_moisture'])


    ####################################################
    ######  SNOW  ######################################
    ####################################################

    #total snow water equivalent in mm
    if cc(config,'conversion','snow_data_source', value='DWD'):
        data['SWE'] = exdata.meteoservice.retrieve_dwd_data(data,lon = lon,
                                                            lat = lat,
                                                            dwd_category='water_equiv',
                                                            dwd_parameters=['total_snow_water_equivalent'],
                                                            no_of_nearest_stations = 5,
                                                            temporal_resolution='daily')['total_snow_water_equivalent']


    ####################################################
    ######  RAIN  ######################################
    ####################################################
    #first we determine the temporal resolution
    #Estimate the temporal resolution for the dwd data in seconds
    temporal_resolutions=np.array([600,3600,86400])
    #get the column with the temporal resolution
    print(tres, 'column determines the temporal resolution')
    #get the closest resolution in seconds
    temporal_resolution=temporal_resolutions.flat[np.abs(temporal_resolutions - data[tres].mean()).argmin()]
    #determine the temporal resolution string for dwd
    if temporal_resolution==600:
        dwd_resolution='10_minutes'
        dwd_par='precipitation_heigth_of_the_last_10_minutes'
        dwd_category='precipitation'
    if temporal_resolution==3600:
        dwd_resolution='hourly'
        dwd_par='hourly_precipitation_height'
        dwd_category='precipitation'
    if temporal_resolution==86400:
        dwd_resolution='daily'
        dwd_category='more_precip'
        dwd_par='daily_precipitation_height'
    print('Temporal Resolution to retrieve DWD data is ',dwd_resolution )
    #add information on current rainfall from station precipitation_heigth_of_the_last_10_minutes
    if cc(config,'conversion','rainfall_data_source', value='DWD'):
        data['rain'] = exdata.meteoservice.retrieve_dwd_data(data,lon = lon,
                                                            lat = lat,
                                                            dwd_category=dwd_category,
                                                            dwd_parameters=[dwd_par],
                                                            no_of_nearest_stations = 5,
                                                            temporal_resolution=dwd_resolution)[dwd_par]


    neutron_columns = [Nfinal]
    if Nth: neutron_columns += [N+th, N+th+clean, N+th+clean+pih]

    ####################################################
    ######  Aggregate ##################################
    ####################################################

    if cc(config,'clean','aggregate'):

        # keep a copy
        data_orig = data.copy()

        # Aggregate
        data = aggregate(data,  gc(config,'clean','aggregate'),
                         func = gc(config,'clean','aggregate_func', alt='mean'),
                         set_std = neutron_columns, adapt_err = neutron_columns,
                         verbose = True)
        # Report
        report_N(data, Nfinal)

    else:
        data_orig = data

    ####################################################
    ######  Smooth  ####################################
    ####################################################

    if cc(config,'clean','smooth'):
        print('| Smooth by a moving average over '+config['clean']['smooth']+' time steps...')

        data = moving_avg(data, Nfinal, window=config['clean']['smooth'])
        Nfinal += mov
        report_N(data, Nfinal)

        #window = int()
        #rollobj = data[N+clean+pih].rolling(window, center=True)
        #data[N+clean+pih+mov]     = rollobj.mean()
        #data[N+clean+pih+mov+std] = rollobj.std()
        #data[N+clean+pih+mov+err] = data[N+clean+pih+err] * 1/np.sqrt(window)

        if Nth:
            data = moving_avg(data, Nthfinal, window=config['clean']['smooth'])
            Nthfinal += mov

            #rollobj = data[N+th+clean+pih].rolling(window, center=True)
            #data[N+th+clean+pih+mov]     = rollobj.mean()
            #data[N+th+clean+pih+mov+std] = rollobj.std()
            #data[N+th+clean+pih+mov+err] = data[N+th+clean+pih+err] * 1/np.sqrt(window)

    if cc(config,'clean','smooth_radius'):

        smooth_radius = gc(config,'clean','smooth_radius', dtype=float)
        _data = data[['utmx','utmy',Nfinal,Nfinal+err]].copy()
        num_neighbors = []

        for i, row in _data.iterrows():
            _data['distance'] = np.sqrt((_data.utmx-row['utmx'])**2+(data.utmy-row['utmy'])**2)

            _query = (_data.distance < smooth_radius) & (~np.isnan(_data[Nfinal]))
            _d = _data[_query]

            if len(_d)==0:
                continue
            elif len(_d)==1:
                _d['distance_weight'] = 1.0

            elif cc(config,'clean','smooth_radius_method','inverse'):
                _d.loc[i,'distance'] = 1.0
                _d['distance_weight'] = 1/_d['distance']

            elif cc(config,'clean','smooth_radius_method','W_r_approx'):
                _d.loc[i,'distance'] = 0.1
                _d['distance_weight'] = _d['distance']*Wr_approx(_d['distance'])

            elif cc(config,'clean','smooth_radius_method','exp'):
                _d['distance_weight'] = np.exp(-_d['distance']/127)

            elif cc(config,'clean','smooth_radius_method','linear'):
                _d['distance_weight'] = 1 - _d['distance']/smooth_radius

            else:
                _d['distance_weight'] = 1.0

            #print(_d)
            data.loc[i, Nfinal+mov]     = np.average(_d[Nfinal], weights=_d['distance_weight'])
            data.loc[i, Nfinal+mov+err] = np.average(_d[Nfinal+err], weights=_d['distance_weight']) / np.sqrt(_d[Nfinal+err].count())
            num_neighbors.append(_d[Nfinal].count())

        print('i Smooth by radius over %d m (median %d neighbors per data point).' % (smooth_radius, np.median(num_neighbors)))
        Nfinal += mov
        report_N(data, Nfinal)

    if not Nfinal.endswith(mov):
        mov = ''

    ####################################################
    ######  Footprint length ###########################
    ####################################################

    if 'Speed_kmh' in data.columns:
        tres = data.index.to_series().dropna().diff().median().total_seconds()
        data['footprint_length'] = data['Speed_kmh']/3.6 * tres
        print('i   Average footprint length is %.0f m.' % data.loc[data['footprint_length'] > 2.0, 'footprint_length'].mean())


    ##JJ smooth with nearest neighbors
    if config['clean']['number_locations']:
        print('| Smooth data with '+config['clean']['number_locations']+' nearest neighbors.')
        data = data.dropna(subset=[lat, lon]) # drop rows without Latitude or Longitude information
        maxdist, RecNums = Neighbours(data[lat], data[lon], int(config['clean']['number_locations']), data)

        data['maxdistance'] = maxdist
        data['neighbors_RecordNum'] = RecNums.tolist()

        test = []
        Recind = []
        res = []
        result = []
        resTh = []
        result_Th = []

        for i in range(len(data)): # store the nearest recordnumber in neighbors
            test = data['RecordNum'].iloc[data['neighbors_RecordNum'][i]].values
            Recind.append(test.tolist())

        data['neighbors_RecordNum'] = Recind

        for i in range(len(data)): # calculate average N from neighbors index
            res = data[Nfinal][data['RecordNum'].isin(data['neighbors_RecordNum'][i])].mean()
            result.append(res)
            if Nth:
                resTh = data[Nthfinal][data['RecordNum'].isin(data['neighbors_RecordNum'][i])].mean()
                result_Th.append(resTh)

        Nfinal += '_neighbors'
        Nthfinal += '_neighbors'
        data[Nfinal] = result
        data[Nthfinal] = result_Th


    ####################################################
    ######  Select Period  #############################
    ####################################################

    data = cut_period_to(data, gc(config,'clean','start'), gc(config,'clean','end'))
    data_orig = cut_period_to(data_orig, gc(config,'clean','start'), gc(config,'clean','end'))

#    if config['clean']['start'] or config['clean']['end']:
#        if config['clean']['start']:
#            start = dateutil.parser.parse(config['clean']['start']).replace(tzinfo=pytz.UTC)
#        else:
#            start = data.index.min()
#        if config['clean']['end']:
#            end = dateutil.parser.parse(config['clean']['end']).replace(tzinfo=pytz.UTC)
#        else:
#            end = data.index.max()
#
#        # cut
#        data = data.loc[start:end]
#        # cut also the raw data
#        data_orig = data_orig.loc[start:end]

    # guess an original data resolution to plot. #TODO Make this configurable later.
    tres_temp = data_orig.index.to_series().dropna().diff().median().total_seconds()
    data_orig = aggregate(data_orig,  gc(config,'clean','aggregate_minor_vis', '1H'),
                         func = gc(config,'clean','aggregate_func', alt='mean'))

    start_str = data.index.min().strftime('%Y%m%d%H')
    end_str = data.index.max().strftime('%Y%m%d%H')
    end_str = '-'+end_str if start_str != end_str else ''
    output_basename = os.path.join(config['output']['out_path'],'') + config['output']['out_basename'] + '-' \
                      + start_str + end_str #config['input']['prefix']

    ####################################################
    ######  Soil Moisture  #############################
    ####################################################

    if config['conversion']['N2sm_method'] != 'None':

        # Calibrate
        N0 = config['conversion']['N0']
        calibration = False
        if not N0:
            print('| Calibration... ', end='')
            calibration = True
            start = dateutil.parser.parse(config['conversion']['campaign_start']).replace(tzinfo=pytz.UTC)
            end = dateutil.parser.parse(config['conversion']['campaign_end']).replace(tzinfo=pytz.UTC)
            datac = data.copy()
            datac = datac.loc[start:end]
            Nc = datac[Nfinal].mean()
            Nc_err = datac[Nfinal].std()
            print('(sm=%.3f, N=%i, bd=%.3f, lw=%.3f, owe=%.3f)' % (float(config['conversion']['measured_sm']), Nc, data.bd.mean(), data.lw.mean(), data.owe.mean()))
            N0     = Calibrate_N0_Desilets(Nc, float(config['conversion']['measured_sm']), bd=data.bd.mean(), lw=data.lw.mean(), owe=data.owe.mean(),
                    a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))       #                Nc/(float(config['conversion']['a0']) / (float(config['conversion']['measured_sm'])/float(config['conversion']['bulk_density']) + float(config['conversion']['a2']) + float(config['conversion']['lattice_water']) + float(config['conversion']['soil_org_carbon'])*0.556) + float(config['conversion']['a1']))
            N0_err = Calibrate_N0_Desilets(Nc+Nc_err, float(config['conversion']['measured_sm']), bd=data.bd.mean(), lw=data.lw.mean(), owe=data.owe.mean(),
                    a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2'])) - N0  #/(float(config['conversion']['a0']) / (float(config['conversion']['measured_sm'])/float(config['conversion']['bulk_density']) + float(config['conversion']['a2']) + float(config['conversion']['lattice_water']) + float(config['conversion']['soil_org_carbon'])*0.556) + float(config['conversion']['a1']))

            print(cprintf('i   N0 = %.0f +/- %.0f cph' % (N0, abs(N0_err)), 'cyan')) # todo: make a report using N0 as column
        else:
            N0 = float(N0)

        # Conversion
        sm = config['conversion']['new_moisture_column']
        sm_neighbors = sm+'_neighbors'

        #while True:
        smv = sm+'_vol'
        smg = sm+'_grv'

        print('| Conversion to water content (N0=%i, bd=%.3f, lw=%.3f, owe=%.3f)... ' % (N0, data.bd.mean(), data.lw.mean(), data.owe.mean()))

        if cc(config,'conversion','N2sm_method','Desilets et al. (2010)'):

            data[smg] = N2SM_Desilets(data[Nfinal], N0, bd=1, lw=data.lw, owe=data.owe,
                        a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))
            data[smv] = N2SM_Desilets(data[Nfinal], N0, bd=data.bd, lw=data.lw, owe=data.owe,
                        a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))
            # same for orig
            data_orig[smv] = N2SM_Desilets(data_orig[N+clean+pih], N0, bd=data_orig.bd, lw=data_orig.lw, owe=data_orig.owe,
                        a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))

        elif cc(config,'conversion','N2sm_method','Schmidt et al. (2021)'):
            data[smv] = N2SM_Schmidt(data, Nfinal, ah, N0, bdstr='bd', lwstr='lw', owestr='owe', method='Mar21_uranos_drf')
            data[smg] = data[smv]/data.bd
            # same for orig
            print('.')
            data_orig[smv] = N2SM_Schmidt(data_orig, Nfinal, ah, N0, bdstr='bd', lwstr='lw', owestr='owe')


        data[smg+err]        = dtheta_stdx(data[Nfinal], data[Nfinal+err], N0, a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))
        data[smg+err+'_low'] = dtheta_low(data[Nfinal], data[Nfinal+err], N0, a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))
        data[smg+err+'_upp'] = dtheta_upp(data[Nfinal], data[Nfinal+err], N0, a0=float(config['conversion']['a0']), a1=float(config['conversion']['a1']), a2=float(config['conversion']['a2']))
        #report_SM(data, smg)

        data['bd_err'] = 0
        if 'bulk_density_err' in config['conversion']:
            data['bd_err'] = float(config['conversion']['bulk_density_err'])
        data[smv+err]        = errpropag_factor(data, 'bd', smg, err=err)                  #data.bd * data[smg+err]        + data[smg]        *
        data[smv+err+'_low'] = errpropag_factor(data, 'bd', smg, err=err, err2=err+'_low') #data.bd * data[smg+err+'_low'] + data[smg+err+'_low'] * float(config['conversion']['bulk_density_err'])
        data[smv+err+'_upp'] = errpropag_factor(data, 'bd', smg, err=err, err2=err+'_upp') #data.bd * data[smg+err+'_upp'] + data[smg+err+'_upp'] * float(config['conversion']['bulk_density_err'])
        if data[smv+err+'_low'].mean() > 0:
            data[smv+err+'_low'] *= -1
        #report_SM(data, smv)

    ####################################################
    ######  Clean soil moisture  #######################
    ####################################################

        validlinesg = len(data[smg].dropna())
        validlinesv = len(data[smv].dropna())
        (a,b) = xsplit(config['clean']['sm_range'], type=float)

        if config['clean']['invalid_data'] == 'drop':
            data = data.query("@"+smg+" > %s and @"+smg+" < %s" % (a,b))
            data = data.query("@"+smv+" > %s and @"+smv+" < %s" % (a,b))
        else:
            data.loc[(data[smg] < a) | (data[smg] > b), smg+err] = np.nan
            data.loc[(data[smg] < a) | (data[smg] > b), smg+err+'_low'] = np.nan
            data.loc[(data[smg] < a) | (data[smg] > b), smg+err+'_upp'] = np.nan
            data.loc[(data[smg] < a) | (data[smg] > b), smg] = np.nan
            data.loc[(data[smg]-data[smg+err] < a), smg+err] = data[smg]
            data.loc[(data[smg]+data[smg+err] > 1), smg+err] = 1-data[smg]
            data.loc[(data[smg]+data[smg+err+'_low'] < a), smg+err+'_low'] = -data[smg]
            data.loc[(data[smg]+data[smg+err+'_upp'] > b), smg+err+'_upp'] = 1-data[smg]

            data.loc[(data[smv] < a) | (data[smv] > b), smv+err] = np.nan
            data.loc[(data[smv] < a) | (data[smv] > b), smv+err+'_low'] = np.nan
            data.loc[(data[smv] < a) | (data[smv] > b), smv+err+'_upp'] = np.nan
            data.loc[(data[smv] < a) | (data[smv] > b), smv] = np.nan
            data.loc[(data[smv]-data[smv+err] < a), smv+err] = data[smv]
            data.loc[(data[smv]+data[smv+err] > b), smv+err] = 1-data[smv]
            data.loc[(data[smv]+data[smv+err+'_low'] < a), smv+err+'_low'] = -data[smv]
            data.loc[(data[smv]+data[smv+err+'_upp'] > b), smv+err+'_upp'] = 1-data[smv]
            #if sm_neighbors in data.columns:
            #    data.loc[(data[sm_neighbors] < a) | (data[sm_neighbors] > b), sm_neighbors] = np.nan

        data[sm] = data[smg]
        data[sm+'_%'] = data[sm] * 100

    #print('dropped %i smg values out of %i' % (validlinesg-len(data[smg].dropna()), validlinesg))
    print('i Dropped %i out of %i values (out of range) from %s.' % (validlinesv-len(data[smv].dropna()), validlinesv, smv))

    report_SM(data, smg, units='grv.%%')
    report_SM(data, smv, units='vol.%%')
    #print('%.3f +/- %.3f -%.3f +%.3f' % (data[smv].mean(), data[smv+err].mean(), data[smv+err+'_low'].mean(), data[smv+err+'_upp'].mean()))

    ####################################################
    ######  Depth  #####################################
    ####################################################

    data['footprint_depth'] = D86(data[smv], data.bd)
    data['footprint_depth'] = data['footprint_depth'].replace(np.inf, np.nan)
    print('i   Footprint depth  D = %3.0f +/- %2.0f cm.' % (data.footprint_depth.dropna().mean(), data.footprint_depth.dropna().std()))

    ####################################################
    ######  Footprint  #################################
    ####################################################

    footprint_data = np.loadtxt(os.path.join(pasdy_path ,'corny/footprint_radius.csv'))
    def get_footprint(sm, h, p):
        if np.isnan(sm) or np.isnan(h) or np.isnan(p): return(np.nan)
        if sm <  0.01: sm =  0.01
        if sm >  0.49: sm =  0.49
        if h  > 30   : h  = 30
        #print(sm, h, p, int(round(100*sm)), int(round(h)))
        return(footprint_data[int(round(100*sm))][int(round(h))] * 0.4922/(0.86-np.exp(-p/1013.25)))

    data['footprint_radius'] = data.apply(lambda row: get_footprint( row[smv], row[ah], row[p] ), axis=1)
    print('i   Footprint radius R = %3.0f +/- %2.0f m.' % (data.footprint_radius.dropna().mean(), data.footprint_radius.dropna().std()))

    data['footprint_volume'] = (data['footprint_depth']+D86(data[smv], data.bd,data['footprint_radius']))*0.01*0.47*data['footprint_radius']**2*3.141 /1000
    print('i   Footprint volume V = %3.0f +/- %2.0f x10^3 m^3.' % (data.footprint_volume.dropna().mean(), data.footprint_volume.dropna().std()))

    # 0.44 (dry) ..0.5 (wet) is roughly the average D over radii

    ####################################################
    ######  Time Resolution  ###########################
    ####################################################

    tres_orig = data_orig.index.to_series().dropna().diff().median().total_seconds()
    tres_orig_str = ''
    if   tres_orig/86400 >= 1: tres_orig_str = '%.0fd' % (tres_orig/86400)
    elif tres_orig/3600  >= 1: tres_orig_str = '%.0fh' % (tres_orig/3600)
    elif tres_orig/60    >= 1: tres_orig_str = '%.0fmin' % (tres_orig/60)
    else                     : tres_orig_str = '%.0fsec' % (tres_orig)

    tres = data.index.to_series().dropna().diff().median().total_seconds()
    tres_str = ''
    if   tres/86400 >= 1: tres_str = '%.0fd' % (tres/86400)
    elif tres/3600  >= 1: tres_str = '%.0fh' % (tres/3600)
    elif tres/60    >= 1: tres_str = '%.0fmin' % (tres/60)
    else                : tres_str = '%.0fsec' % (tres)

    ####################################################
    ######  POI  #######################################
    ####################################################

    if cc(config,'conversion','poi_table'):
        print('| Evaluating points of interest...')
        poi = '_poi'

        #POI = read_geojson(config['conversion']['poi_json']) # 'data/Selke.geojson'
        poi_list = gc(config,'conversion','poi_table', dtype='multi')

        _poi_data_columns = ['name','lat','lon','radius','utmx','utmy',Nfinal,Nfinal+err,Nfinal+std,smv,smv+err,smv+std]
        _poi_data = []
        for item in poi_list:
            _poi_data.append(xsplit(item, range=len(_poi_data_columns)))
        poi_data = pandas.DataFrame(_poi_data, columns=_poi_data_columns)
        del _poi_data

        poi_data.lat = pandas.to_numeric(poi_data.lat)
        poi_data.lon = pandas.to_numeric(poi_data.lon)
        poi_data.radius = pandas.to_numeric(poi_data.radius)
        utmy, utmx = latlon2utm(poi_data.lat, poi_data.lon)
        poi_data['utmy'] = utmy
        poi_data['utmx'] = utmx

        _data = data[['utmx','utmy',Nfinal,Nfinal+err,smv,smv+err]].copy()

        for i, row in poi_data.iterrows():
            _data['distance2poi'] = np.sqrt((_data.utmx-row['utmx'])**2+(data.utmy-row['utmy'])**2)

            if cc(config,'conversion','poi_distance_weight','inverse'):
                _data['distance_weight'] = 1/_data['distance2poi']
            elif cc(config,'conversion','poi_distance_weight','W_r_approx'):
                _data['distance_weight'] = Wr_approx(_data['distance2poi'])
            else:
                _data['distance_weight'] = 1

            _query = (_data.distance2poi < row['radius']) & (~np.isnan(_data[Nfinal]))
            poi_data.loc[i, Nfinal]     = np.average(_data.loc[_query, Nfinal],     weights=_data.loc[_query, 'distance_weight'])
            poi_data.loc[i, Nfinal+err] = np.average(_data.loc[_query, Nfinal+err], weights=_data.loc[_query, 'distance_weight']) / np.sqrt(_data.loc[_query, Nfinal].count())
            poi_data.loc[i, Nfinal+std] =            _data.loc[_query, Nfinal].std()
            _query = (_data.distance2poi < row['radius']) & (~np.isnan(_data[smv]))
            poi_data.loc[i, smv]        = np.average(_data.loc[_query, smv],        weights=_data.loc[_query, 'distance_weight'])
            poi_data.loc[i, smv+err]    = np.average(_data.loc[_query, smv+err],    weights=_data.loc[_query, 'distance_weight']) / np.sqrt(_data.loc[_query, smv].count())
            poi_data.loc[i, smv+std]    =            _data.loc[_query, smv].std()

            print('|   POI %15s: N = %5.0f +/-%4.0f (%4.0f) cph, SM = %4.1f +/-%4.1f (%4.1f) vol.%%' % (row['name'],
                poi_data.loc[i, Nfinal],  poi_data.loc[i, Nfinal+err],  poi_data.loc[i, Nfinal+std],
                poi_data.loc[i, smv]*100, poi_data.loc[i, smv+err]*100, poi_data.loc[i, smv+std]*100))

        poi_data.to_csv(output_basename+'_POI.csv',
            sep = config['output']['CSV_column_sep'].strip('\"'),
            decimal = config['output']['CSV_decimal'],
            float_format = '%' + config['output']['CSV_float_format'],
            na_rep = config['output']['CSV_NaN'])
        file_save_msg(output_basename + '_POI.csv', "POI", end="\n")
    else:
        poi = ''


    ####################################################
    ######  Output  ####################################
    ####################################################

    if yesno2bool(config['output']['make_CSV']):
        print('| CSV data preparation... ')
        datetime_format = re.sub(r'(\w)', r'%\1', config['output']['CSV_datetime_format'])
        if cc(config,'output','CSV_columns'):
            out_columns = []
            column_names = xsplit(gc(config,'output','CSV_columns'))
            for item in column_names:
                item = item.replace('neutrons_proc', Nfinal)
                item = item.replace('neutrons_raw', N)
                item = item.replace('relative_humidity', rh)
                item = item.replace('temperature', T)
                item = item.replace('pressure', p)
                item = item.replace('alt', 'Alt')
                if item in data.columns:
                    out_columns.append(item)
        else:
            out_columns  = data.columns
            column_names = data.columns

        try:
            data[out_columns].to_csv(output_basename+'.csv',
                date_format = datetime_format, header=column_names,
                sep = config['output']['CSV_column_sep'].strip('\"'),
                decimal = config['output']['CSV_decimal'],
                float_format = '%' + config['output']['CSV_float_format'],
                na_rep = config['output']['CSV_NaN'])
        except:
            print('! ERROR: Cannot write, probably permission denied.')

        file_save_msg(output_basename + '.csv', "Data", end="\n")


    # Figures
    if yesno2bool(config['output']['make_PDF']):

        print('| Plotting... ', end='')
        figure_name = output_basename + '.pdf'

        #print(data[N].to_string())
        ## With PdfPages: Create PDF on enter and close PDF on exit
        ## see https://matplotlib.org/gallery/misc/multipage_pdf.html
        with matplotlib.backends.backend_pdf.PdfPages(figure_name) as fig:

            Fig.time_format = config['output']['PDF_time_format']

            # Cover
            if 'info' in config:

                stamp = 'CoRNy v%s\n%s' % (VERSION, datetime.now().strftime('%b %d, %Y\n%H:%M UTC'))
                with Fig(fig, title='Data Report', size=(10.4,5.5)) as ax:
                    plt.axis('off')
                    plt.annotate(stamp, (0.99,0.99), fontsize=7, ha='right')
                    plt.annotate(config['info']['title'].encode('latin-1').decode('utf-8'), (0.05,0.9), fontsize=16)
                    plt.annotate(config['info']['subtitle'].encode('latin-1').decode('utf-8'), (0.05,0.82), fontsize=9)
                    plt.annotate('Contact: '+config['info']['contact'].encode('latin-1').decode('utf-8'), (0.05,0.78), fontsize=9)
                    #import matplotlib.image as mpimg
                    if os.path.exists(config['info']['cover']):
                        logo = plt.imread(config['info']['cover'])
                        ax.figure.figimage(logo, 0.3, 0.5, zorder=1)

            # Neutrons, label='%s resolution, $\pm$%0.f smoothing window' % (tres_str, int(config['clean']['smooth'])/2))
            if config['clean']['smooth']:
                title_str = 'Neutron time series (%s aggregation, $\pm$%0.f smoothing window)' % (tres_str, int(config['clean']['smooth'])/2)
            else:
                title_str = 'Neutron time series (%s aggregation)' % (tres_str)

            with Fig(fig, title=title_str, ylabel='Neutrons (in cph)') as ax:
                if not data[N].isnull().all():
                    plt.plot(data[N],           label="N raw counts in cph",     ls='', marker='o', ms=3, markeredgewidth=0, color='red')
                    plt.plot(data[N+clean],     label="N cleaned", ls='', marker='o', ms=3.5, markeredgewidth=0, color='whitesmoke')
                    #plt.plot(data[N+clean+pih], label="N_cc_pih", ls='', marker='o', ms=3,  color='C0', alpha=0.6, markeredgewidth=0) # , drawstyle='steps'
                    #plt.plot(data[N+clean+pih+urb+bio], label="N_cc_pih_ub", ls='', marker='o', ms=3,  color='C1', alpha=0.6, markeredgewidth=0)
                    if Nfinal == N+clean+pih:
                        plt.plot(data[N+clean+pih],                                lw=2, label="N c&corr press,incom,humid", color='black')
                    else:
                        plt.plot(data[N+clean+pih], ls='', marker='o', ms=3.5, markeredgewidth=0, label="N c&corr press,incom,humid", color='lightgrey')
                    #if urb != '': plt.plot(data[N+clean+pih+urb],              lw=4, label="N c&corr press,incom,humid,urban", color='C6')
                    #if bio != '': plt.plot(data[N+clean+pih+urb+bio],          lw=3, label="N c&corr press,incom,humid,urban,bio", color='C2')
                    #if road!= '': plt.plot(data[N+clean+pih+urb+bio+road],     lw=2, label="N c&corr press,incom,humid,urban,bio,road", color='C3')
                    if mov != '': plt.plot(data[N+clean+pih+urb+bio+road+mov], lw=1, label="N c&corr press,incom,humid,urban,bio,road,smoothed", color='black')
                    #if Nfinal != N+clean+pih:
                    #    plt.plot(data[Nfinal], label="N cleaned corrected_pih smoothed_%s" % config['clean']['smooth'], color='C0')
                    ax.fill_between(data.index, data[Nfinal]-data[Nfinal+'_err'], data[Nfinal]+data[Nfinal+'_err'], color='black', alpha=0.2, lw=0)
                    #plot_errorbar(data.index, data[Nfinal], xerr=None, yerr=data[Nfinal+err], fmt='o', ecolor='lightgrey', elinewidth=1, capsize=0, mfc='black', ms=3, mew=0)

                else:
                    plt.plot(data.index[0], 0)

                if 'N_neighbors' in data.columns:
                    #JJ
                    if not data['N_neighbors'].isnull().all():
                        plt.plot(data['N_neighbors'], label='N_neighbors: spatially smoothed', drawstyle='steps')

                plt.legend()

            if Nth:

                with Fig(fig, title='Thermal neutrons', ylabel='Thermal Neutrons (in cph)') as ax:
                    if not data[N].isnull().all():
                        plt.plot(data[N+th+clean],     label="N_thermal cleaned", ls='', marker='o', ms=3.5, markeredgewidth=0, color='whitesmoke')
                        plt.plot(data[N+th+clean+pih], label="N_thermal c&corr press,incom,humid", lw=2, color='black')
                        plt.legend()

                with Fig(fig, title='Thermal and Epithermal neutrons corrected for p,i,h (sea level)', ylabel='Neutrons (in cph)') as ax:
                    if not data[N].isnull().all():
                        plt.plot(data[N+clean+pih], label="Epithermal", lw=2, color='C0')
                        plt.plot(data[N+th+clean+pih],     label="Thermal", lw=2, color='C1')
                        plt.legend()

        #            if N+'_std' in data.columns:
        #                ax.fill_between(data.index, data[N]-data[N+'_std'], data[N]+data[N+'_std'], alpha=0.25, lw=0, step='pre')

#            if Nth:
#                with Fig(fig, title='Thermal neutron time series, corrected', ylabel='Thermal neutrons (in cph)') as ax:
#                    if not data[Nth].isnull().all():
#                        plt.plot(data[Nth], label="N_thermal", drawstyle='steps')


            # Corrections
            with Fig(fig, title='Corrections', ylabel='-') as ax:
                if not data[N].isnull().all():
                    plt.plot(data['correct_p'], label='air pressure', color='grey')
                    plt.plot(data['correct_i'], label='incoming neutron flux', color='C1')
                    plt.plot(data['correct_h'], label='air humidity', color='C0')
                    if urb != '': plt.plot(data['correct_urban'], label='urban area', lw=5, color='C6')
                    if bio != '': plt.plot(data['correct_bio'], label='biomass', color='C2')
                    if road!= '': plt.plot(data['correct_road'], label='road effect', color='black', lw=1)
                    plt.legend()
                else:
                    plt.plot(data.index[0], 0)

                if config['conversion']['road_method'] != '' and config['conversion']['road_method'] != 'None'and lat in data.columns:
                    with Fig(fig, title='Road parameters', ylabel='-') as ax:
                        plt.plot(data['road_width'], label='road_width', marker='o', ls='', ms=2)
                        plt.plot(data['road_moisture'], label='road_moisture', marker='o', ls='', ms=2)
                        plt.legend()

            if data.bd.std().mean() > 0.001:
                with Fig(fig, title='Soil properties', ylabel='-') as ax:
                    plt.plot(data['bd'], label='bulk density', marker='o', ls='', ms=2)
                    plt.plot(data['lw'], label='lattice water', marker='o', ls='', ms=2)
                    plt.plot(data['owe'], label='Organic water equivalent', marker='o', ls='', ms=2)
                    plt.legend()

            # Air properties
            with Fig(fig, layout=(2,2)) as axes:
                if not data[p].isnull().all():
                    with Fig(ylabel='Air pressure (in mbar)'):
                        plt.plot(data[p], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=1: axes[1].axis('off')

                if not data[T].isnull().all():
                    with Fig(ylabel='Air temperature (in $^\circ$C)'):
                        plt.plot(data[T], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=2: axes[2].axis('off')

                if not data[rh].isnull().all():
                    with Fig(ylabel='Relative humidity (in %)'):
                        plt.plot(data[rh], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=3: axes[3].axis('off')

                if not data[ah].isnull().all():
                    with Fig(ylabel='Absolute humidity (in g/m$^3$)'):
                        plt.plot(data[ah], drawstyle='steps')
                else:
                    pass#if len(list(axes))>=4: axes[4].axis('off')

            # Moisture
            if config['conversion']['N2sm_method'] != 'None':
                if not data[sm].isnull().all():


#                    with Fig(fig, title='Water content time series, corrected', ylabel='Gravimetric water content', ylim=(0,1)) as ax:
#                        plt.plot(data[smg], color='black', ls='', marker='o', ms=4, markeredgewidth=0)
                        #ax.fill_between(data.index, data[smg]+data[smg+err+'_low'], data[smv]+data[smg+err+'_upp'], color='C0', alpha=0.4, lw=0, step='pre')
#                        plot_errorbar(data.index, data[smg], xerr=None, yerr=[-data[smg+err+'_low'], data[smg+err+'_upp']], fmt='o', ecolor='lightgrey', elinewidth=1, capsize=0, mfc='black', ms=3, mew=0)
#                        plt.yticks(np.arange(0,1.1, 0.1))

#                    with Fig(fig, title='Water content time series, corrected', ylabel='Volumetric water content', ylim=(0,1)) as ax:
#                        plt.plot(data[smv], color='black', ls='', marker='o', ms=4, markeredgewidth=0)
                        #ax.fill_between(data.index, data[smv]+data[smv+err+'_low'], data[smv]+data[smv+err+'_upp'], color='C0', alpha=0.4, lw=0, step='pre')
#                        plot_errorbar(data.index, data[smv], xerr=None, yerr=[-data[smv+err+'_low'], data[smv+err+'_upp']], fmt='o', ecolor='lightgrey', elinewidth=1, capsize=0, mfc='None', ms=3, mew=1)
#                        plt.yticks(np.arange(0,1.1, 0.1))
                    if config['clean']['smooth']:
                        label_str = '%s resolution, $\pm$%0.f smoothing window' % (tres_str, int(config['clean']['smooth'])/2)
                    else:
                        label_str = '%s resolution' % (tres_str)

                    with Fig(fig, title='Water content time series, corrected', ylabel='Volumetric water content (in cm$^3$/cm$^3$)', ylim=(0,1)) as ax:
                        ax.fill_between(data.index, data[smv], 0, color='C0', alpha=0.05, lw=0, step='post')
                        plt.plot(data_orig[smv], ls='', color='C0', alpha=0.3, marker='o', ms=2, markeredgewidth=0, label='%s resolution' % tres_orig_str)
                        plt.plot(data[smv], color='C0', drawstyle='steps-post', lw=2, label=label_str)
                        #ax.fill_between(data.index, data[smv]+data[smv+err+'_low'], data[smv]+data[smv+err+'_upp'], color='C0', alpha=0.4, lw=0, step='pre')
                        #plot_errorbar(data.index, data[smv], xerr=None, yerr=[-data[smv+err+'_low'], data[smv+err+'_upp']], fmt='o', ecolor='lightgrey', elinewidth=1, capsize=0, mfc='None', ms=3, mew=1)
                        plt.yticks(np.arange(0,1.1, 0.1))
                        plt.legend(ncol=2, loc='upper center')
                        #ax.set_xlim(str2dt('2020-03-01'),str2dt('2020-09-01'))
                        #plt.scatter(_data.loc[_query, Nfinal].index, _data.loc[_query, smv], color='red')

                    with Fig(fig, title='Error of the volumetric water content', ylabel='Error (in cm$^3$/cm$^3$)', #) as ax:
                        ylim=(-1.1*data[smv+err+'_upp'].max(), 1.1*data[smv+err+'_upp'].max())) as ax:
                        plt.plot(data[smv+err+'_low'], drawstyle='steps', color='C3', lw=1)
                        plt.plot(data[smv+err+'_upp'], drawstyle='steps', color='C2', lw=1)
                        ax.fill_between(data.index, data[smv+err+'_low'], 0, color='C3', alpha=0.2, lw=0, step='pre')
                        ax.fill_between(data.index, data[smv+err+'_upp'], 0, color='C2', alpha=0.2, lw=0, step='pre')

#                        data[smv+'_int'] = data[smv].interpolate()
#                        data[smv+err+'_low_int'] = data[smv+err+'_low'].interpolate()
#                        data[smv+err+'_upp_int'] = data[smv+err+'_upp'].interpolate()
#                        data = moving_avg(data, smv+'_int', window=15, err=False, std=False)
#                        data = moving_avg(data, smv+err+'_low_int', window=15, err=False, std=False)
#                        data = moving_avg(data, smv+err+'_upp_int', window=15, err=False, std=False)

#                        plt.plot(data[smv+'_int'+mov], color='black', alpha=0.3, drawstyle='steps')
                        #plt.plot(data[smv+'_int'+mov]+data[smv+err+'_low_int'+mov], color='red')
                        #plt.plot(data[smv+'_int'+mov]+data[smv+err+'_upp_int'+mov], color='red')
#                        ax.fill_between(data.index, data[smv+'_int'+mov]+data[smv+err+'_low_int'+mov],
#                                                    data[smv+'_int'+mov]+data[smv+err+'_upp_int'+mov],
#                                                    color='black', alpha=0.1, lw=0, step='pre')

                        #plt.plot(data[smv] +data[smv+err+'_low'], color='C1', ls='', marker='o', ms=2, markeredgewidth=0)
                        #plt.plot(data[smv] +data[smv+err+'_upp'], color='C1', ls='', marker='o', ms=2, markeredgewidth=0)

#                        plt.errorbar(data.index, data[smv], yerr=[data[smv] +data[smv+err+'_low'], data[smv] +data[smv+err+'_upp']],
#                                     fmt='s', ecolor='black', elinewidth=1, capsize=0, mfc='black', mec='black', ms=4, mew=2)

#                    with Fig(fig, title='Water content time series, corrected', ylabel='Gravimetric water content', ylim=(0,1)):
#                        plt.plot(data[smg], color='black', drawstyle='steps')
#                        plt.plot(data[smg] +data[smg+err+'_low'], color='C1', drawstyle='steps')
#                        plt.plot(data[smg] +data[smg+err+'_upp'], color='C1', drawstyle='steps')


#                        ax.fill_between(data.index, data[smv]+data[smv+err+'_low'], data[smv]+data[smv+err+'_upp'], color='C0', alpha=0.4, lw=0, step='pre')
                        #plt.plot(data[smv] -data[smv+err], color='C2')
                        #plt.plot(data[smv] +data[smv+err], color='C2')
#                        if calibration:
#                            plt.plot([datac.index.min(), datac.index.max()], [float(config['conversion']['measured_sm'])*100, float(config['conversion']['measured_sm'])*100], color='red', lw=10, alpha=0.5)
#                            plt.text(datac.index.min(), float(config['conversion']['measured_sm'])*105, 'calibration', rotation=90, va='bottom', color='red', alpha=0.5)

#                    datax = data.copy()
#                    datax[smv] = datax[smv].interpolate()
                    #datax = moving_avg(datax, smv, window=15, err=False, std=False)
#                    datax[smv+err+'_low'] = datax[smv+err+'_low'].interpolate()
                    #datax = moving_avg(datax, smv+err+'_low', window=15, err=False, std=False)
#                    datax[smv+err+'_upp'] = datax[smv+err+'_upp'].interpolate()
                    #datax = moving_avg(datax, smv+err+'_upp', window=15, err=False, std=False)
#                    datay = minly(datax,5)
#                    datay = moving_avg(datay, smv, window=6, err=False, std=False)
#                    datay = moving_avg(datay, smv+err+'_low', window=6, err=False, std=False)
#                    datay = moving_avg(datay, smv+err+'_upp', window=6, err=False, std=False)
#                    datay = moving_avg(datay, smg, window=6, err=False, std=False)
#                    datay = moving_avg(datay, smg+err+'_low', window=6, err=False, std=False)
#                    datay = moving_avg(datay, smg+err+'_upp', window=6, err=False, std=False)

#                    with Fig(fig, title='Water content time series, corrected', ylabel='Volumetric water content', ylim=(0,1)) as ax:
#                        plt.plot(datay[smv+mov], color='black', drawstyle='steps')
                        #plt.plot(datay[smv+mov] +datay[smv+err+'_low'+mov], color='C1', drawstyle='steps')
                        #plt.plot(datay[smv+mov] +datay[smv+err+'_upp'+mov], color='C1', drawstyle='steps')
#                        ax.fill_between(datay.index, datay[smv+mov]+datay[smv+err+'_low'+mov], datay[smv+mov]+datay[smv+err+'_upp'+mov], color='C0', alpha=0.4, lw=0, step='pre')

#                    with Fig(fig, title='Water content time series, corrected', ylabel='Gravimetric water content', ylim=(0,1)) as ax:
#                        plt.plot(datay[smg+mov], color='black', drawstyle='steps')
#                        ax.fill_between(datay.index, datay[smg+mov]+datay[smg+err+'_low'+mov], datay[smg+mov]+datay[smg+err+'_upp'+mov], color='C0', alpha=0.4, lw=0, step='pre')

                    # hist
                    print('Histogram... ', end='')
                    with Fig(fig, title='PDF '+sm, ylabel='probability density', time_series=False, size=(10,5)) as ax:
                        ax.hist(data[sm].dropna().values, bins=100, density=True, histtype='stepfilled')
                        ax.set_xlim(0,1)
                        ax.set_xticks(np.arange(0,1,0.1))
                        ax.set_xlabel('vol. soil moisture (in m$^3$/m$^3$)')
                        #ax.set_ylabel('probability density')
                        #ax.set_title('Soil moisture PDF') ;
                        ax.grid(b=True, color='black', alpha=0.1)


            # More plots
            for var in xsplit(config['output']['PDF_plots']):

                ## Footprint
                if var == 'footprint':
                    print('footprint... ', end='')

                    with Fig(fig, layout=(3,1)) as axes:

                        var = 'footprint_radius'
                        with Fig(title='Footprint', ylabel='Radius (in m)', ylim=(0,350)) as ax:
                            ax.fill_between(data.index, data[var], 0, color='C2', alpha=0.3, lw=0, step='pre')
                            plt.plot(data[var], color='C2', drawstyle='steps', lw=2)
                            #plt.yticks(np.arange(0,1.1, 0.1))
                            #plt.legend(ncol=2, loc='upper center')
                            plt.grid(b=True, color='black', alpha=0.2)

                        var = 'footprint_depth'
                        with Fig(ylabel='Depth (in cm)', ylim=(-80,0)) as ax:
                            ax.fill_between(data.index, -data[var], 0, color='sienna', alpha=0.3, lw=0, step='pre')
                            plt.plot(-data[var], color='sienna', drawstyle='steps', lw=2)
                            #plt.yticks(np.arange(0,1.1, 0.1))
                            #plt.legend(ncol=2, loc='upper center')
                            plt.grid(b=True, color='black', alpha=0.2)

                        var = 'footprint_volume'
                        with Fig(ylabel='Volume (in $10^3$ m$^3$)', ylim=(0,120)) as ax:
                            ax.fill_between(data.index, data[var], 0, color='C0', alpha=0.3, lw=0, step='pre')
                            plt.plot(data[var], color='C0', drawstyle='steps', lw=2)
                            #plt.yticks(np.arange(0,1.1, 0.1))
                            #plt.legend(ncol=2, loc='upper center')
                            plt.grid(b=True, color='black', alpha=0.2)

                elif var == 'tubes':
                    print('tubes... ', end='')

                    N_tubes = xsplit(config['input']['neutron_columns'])
                    mysize = np.ceil(np.sqrt(len(N_tubes)))
                    if len(N_tubes) == 2:
                        mylayout = (2,1)
                    else:
                        mylayout = (int(mysize),int(mysize))
                    ymax = data[N_tubes[1]].mean()*4

                    with Fig(fig, layout=mylayout) as axes:
                        for var in N_tubes:

                            with Fig(ylabel='raw counts', ylim=(0,ymax)) as ax:
                                ax.plot(data[var], color='red', ls='', marker='o', ms=2, markeredgewidth=0)
                                ax.plot(data[var+'_clean'], color='silver', ls='', marker='o', ms=2.1, markeredgewidth=0)

                                rollobj = data[var+'_clean'].interpolate(limit=gc(config,'clean','interpolate_neutrons_gapsize',dtype=int)).rolling(int(config['clean']['smooth']))
                                ax.plot(rollobj.mean(), color='red', ls='-', lw=1)
                                rollobj = data[var+'_clean'].rolling(int(config['clean']['smooth']))
                                ax.plot(rollobj.mean(), color='black', ls='-', lw=2)


                                ax.grid(b=True, color='black', alpha=0.2)
                                ax.text(data[var].index[0], ymax*0.95, var)

                        for i in range(int(mysize)*int(mysize)-len(N_tubes)):
                            with Fig(ylabel='N', ylim=(0,3)) as ax:
                                ax.set_axis_off()
                            #with Fig(ylabel='N', ylim=(0,3)):
                            #    pass #plt.plot(data[N], color='black', ls='', marker='o', ms=2, markeredgewidth=0)


                ## Plot gridded data (EXPERIMENTAL)
                elif var == 'map_grid':

                    var = sm
                    if lat != '' and lat in data.columns:
                        datax = data[[lon,lat,var]].dropna()
                        if len(datax)==0:
                            raise ValueError('No records with valid sm left after processing, stop printing')

                        ## Deprecated:
                        # import pyproj
                        # xs, ys = pyproj.transform(pyproj.Proj("epsg:4326"), pyproj.Proj("epsg:31468"), datax[lat].values, datax[lon].values)
                        ## New pyproj2 syntax:
                        from pyproj import Transformer
                        xs, ys = Transformer.from_crs("epsg:4326", "epsg:31468").transform(datax[lat].values, datax[lon].values)

                        datax['y'] = xs
                        datax['x'] = ys

                        from scipy.interpolate import griddata
                        # data coordinates and values
                        x = datax.x.values - datax.x.min()
                        y = datax.y.values - datax.y.min()
                        z = datax[var].values

                        # config
                        print('XY plot... ', end='')
                        resolution = 70
                        smrange = (0,0.5)
                        contour_levels = np.arange(0.05,0.525,0.025) #20

                        xrange = x.max()-x.min()
                        yrange = y.max()-y.min()
                        if xrange==0: xrange=1
                        if yrange==0: yrange=1
                        xpad = 0.05*xrange
                        ypad = 0.10*yrange

                        mysize = (10,10/xrange*yrange)
                        if mysize[1] > 20:
                            mysize = (10*xrange/yrange,10)
                        with Fig(fig, title='Map '+var, xlabel='Easting (in m)', ylabel='Northing (in m)', size=mysize, time_series=False) as ax:
                            q = ax.scatter( x, y, c=z, cmap='Spectral', vmin=smrange[0], vmax=smrange[1])
                            ax.set_aspect(1)
                            ax.set_xlim(x.min()-xpad,x.max()+xpad)
                            ax.set_ylim(y.min()-ypad,y.max()+ypad)
                            #ax.set_xlabel('Easting [m]')
                            #ax.set_ylabel('Northing [m]')
                            #ax.set_title('Soil Moisture Map') ;
                            cb = plt.colorbar(q, ax=ax, shrink=0.6, pad=0.03, aspect=30)
                            cb.set_label("vol. soil moisture (in m$^3$/m$^3$)", rotation=270, labelpad=20)
                            plt.grid(color='black', alpha=0.1)

                            if cc(config,'conversion','poi_plot','yes'):
                                for i, poi in poi_data.iterrows():
                                    ax.scatter( poi.utmx- datax.x.min(), poi.utmy- datax.y.min(), c='black', marker='+', s=100)
                                    circle = plt.Circle((poi.utmx- datax.x.min(), poi.utmy- datax.y.min()), poi.radius, fc='none', color='black', ls='--')
                                    ax.add_patch(circle)


                ## Plot with tiles (EXPERIMENTAL)
                elif var == 'map_tiles':

                    var = sm
                    # TODO Those params should go to the config file
                    todo_config_zoom = gc(config,'output','satellite_zoom', 8, dtype=int)
                    todo_config_map = 'satellite'
                    todo_config_markersize = 15

                    print('Latlon plot with tiles... ', end='')
                    datall = data.dropna(subset=[sm]).copy()

                    if todo_config_map == 'satellite':
                        cimgt.QuadtreeTiles.get_image = image_spoof # reformat web request for street map spoofing
                        osm_img = cimgt.QuadtreeTiles() # spoofed, downloaded street map
                    elif todo_config_map == 'OSM':
                        cimgt.OSM.get_image = image_spoof # reformat web request for street map spoofing
                        osm_img = cimgt.OSM() # spoofed, downloaded street map
                    else:
                        todo_config_map = None

                    lonpad = 0.1*(datall[lon].max()-datall[lon].min())
                    latpad = 0.3*(datall[lat].max()-datall[lat].min())

                    datax = data[[lon,lat,var]].dropna()
                    if len(datax)==0:
                        raise ValueError('No records with valid sm left after processing, stop printing')

                    ## Deprecated:
                    # import pyproj
                    # xs, ys = pyproj.transform(pyproj.Proj("epsg:4326"), pyproj.Proj("epsg:31468"), datax[lat].values, datax[lon].values)
                    ## New pyproj2 syntax:
                    from pyproj import Transformer
                    xs, ys = Transformer.from_crs("epsg:4326", "epsg:31468").transform(datax[lat].values, datax[lon].values)

                    datax['y'] = xs
                    datax['x'] = ys

                    from scipy.interpolate import griddata
                    # data coordinates and values
                    x = datax.x.values - datax.x.min()
                    y = datax.y.values - datax.y.min()
                    z = datax[var].values

                    smrange = (0,0.5)

                    xrange = x.max()-x.min()
                    yrange = y.max()-y.min()
                    if xrange==0: xrange=1
                    if yrange==0: yrange=1
                    xpad = 0.05*xrange
                    ypad = 0.10*yrange

                    mysize = (10,10/xrange*yrange)
                    if mysize[1] > 20:
                        mysize = (10*xrange/yrange,10)

                    zoom = todo_config_zoom #18-int(15*(data[lon].max()-data[lon].min())) # 0->18, 0.5->16, 1->14
                    # Do not choose >=17 for more than a small field site! Python will crash.
                    #print("(zoom %i) " % zoom, end='')

                    with Fig(fig, title='Map '+var, xlabel='lon', ylabel='lat', size=mysize, time_series=False, proj=ccrs.PlateCarree()) as ax:
                        ax.scatter( datall[lon], datall[lat], c='black', transform=ccrs.PlateCarree(), s=todo_config_markersize*2)
                        q = ax.scatter( datall[lon], datall[lat], c=datall[sm], transform=ccrs.PlateCarree(),
                                        s=todo_config_markersize, cmap='Spectral', vmin=smrange[0], vmax=smrange[1])
                        ax.set_xlim(datall[lon].min()-lonpad, datall[lon].max()+lonpad)
                        ax.set_ylim(datall[lat].min()-latpad, datall[lat].max()+latpad)
                        if not todo_config_map is None:
                            ax.add_image(osm_img, zoom)

                        cb = plt.colorbar(q, ax=ax, shrink=0.7, pad=0.03, aspect=30)
                        cb.set_label("vol. soil moisture (in m$^3$/m$^3$)", rotation=270, labelpad=20)

                        if cc(config,'conversion','poi_plot','yes'):
                            for i, poi in poi_data.iterrows():
                                ax.scatter( poi.lon, poi.lat, c='black', marker='+', s=100)

                ## Plot grid with tiles (EXPERIMENTAL)
                elif var == 'map_grid_tiles':

                    var = sm
                    # TODO Those params should go to the config file
                    todo_config_zoom = gc(config,'output','satellite_zoom',8, dtype=int)
                    todo_config_map = 'satellite'
                    todo_config_resolution = gc(config,'output','grid_resolution',alt=3, dtype=int)
                    todo_config_fineresolu = gc(config,'output','grid_resol_fine',alt=2.1, dtype=float)   # Selbitz 2.1 # 4.5 #1.5*0.62*4 # Corinna, 1.5*0.62 # Hordorf
                    todo_config_markersize = gc(config,'output','grid_markersize',alt=100, dtype=float) # 400 #100/4.5 # Corinna, 100 # Hordorf

                    print('Latlon plot with regular grid and tiles...', end='')

                    if todo_config_map == 'satellite':
                        cimgt.QuadtreeTiles.get_image = image_spoof # reformat web request for street map spoofing
                        osm_img = cimgt.QuadtreeTiles() # spoofed, downloaded street map
                    elif todo_config_map == 'OSM':
                        cimgt.OSM.get_image = image_spoof # reformat web request for street map spoofing
                        osm_img = cimgt.OSM() # spoofed, downloaded street map
                    else:
                        todo_config_map = None

                    #print(cprintf('! Internet connection failed. Skipping basemap...','red'))

                    zoom = todo_config_zoom #18-int(15*(data[lon].max()-data[lon].min())) # 0->18, 0.5->16, 1->14
                    # Do not choose >=17 for more than a small field site! Python will crash.
                    #print("(zoom %i) " % zoom, end='')

                    G = data[[lat,lon,smv,rh,T]]
                    G[smv+'_pc'] = G[smv]*100
                    G.loc[:,lat] = fineround(G[lat], todo_config_resolution, todo_config_fineresolu)
                    G.loc[:,lon] = fineround(G[lon], todo_config_resolution, todo_config_fineresolu)
                    G = G.groupby([lat,lon]).mean()
                    G = G.reset_index() # from MultiIndex back to normal columns

                    utmy, utmx = latlon2utm(G[lat], G[lon])
                    G['utmx'] = utmx
                    G['utmy'] = utmy
                    G.to_csv(output_basename+'-grid.csv',
                        date_format = datetime_format,
                        sep = config['output']['CSV_column_sep'].strip('\"'),
                        decimal = config['output']['CSV_decimal'],
                        float_format = '%' + config['output']['CSV_float_format'],
                        na_rep = config['output']['CSV_NaN'])

                    lonpad = 0.1*(data[lon].max()-data[lon].min())
                    latpad = 0.3*(data[lat].max()-data[lat].min())

                    with Fig(fig, title='CRNS aggregated map '+var, size=mysize, time_series=False, proj=ccrs.PlateCarree()) as ax:

                        q = ax.scatter( G[lon], G[lat], c=G[smv], s=todo_config_markersize, marker='o', edgecolor='black', linewidth=0.5,
                                       cmap='Spectral', vmin=smrange[0], vmax=smrange[1], transform=ccrs.PlateCarree())
                        ax.scatter( data[lon], data[lat], c='black', s=2, transform=ccrs.PlateCarree(), edgecolor='none', linewidth=0)
                        ax.set_xlim(data[lon].min()-lonpad, data[lon].max()+lonpad)
                        ax.set_ylim(data[lat].min()-latpad, data[lat].max()+latpad)
                        #ax.set_xticks(np.arange(data[lon].min(), data[lon].max(), lonpad))
                        ax.set_yticks(np.around(np.arange(bbox[2], bbox[3], (bbox[3]-bbox[2])/5),4))
                        ax.set_xticks(np.around(np.arange(bbox[0], bbox[1], (bbox[1]-bbox[0])/5),4))
                        ax.ticklabel_format(axis='both', style='plain', scilimits=[-10,6], useOffset=False, useMathText=True)

                        if not todo_config_map is None:
                            ax.add_image(osm_img, zoom)

                        cb = plt.colorbar(q, ax=ax, shrink=0.5, pad=0.03, aspect=30)
                        cb.set_label("vol. soil moisture (in m$^3$/m$^3$)", rotation=270, labelpad=20)
                        plt.grid(color='black', alpha=0.1)

                        if cc(config,'conversion','poi_plot','yes'):
                            for i, poi in poi_data.iterrows():
                                ax.scatter( poi.lon, poi.lat, c='black', marker='+', s=100)

                    if cc(config, 'output', 'map_grid_tiles_temperature', 'yes'):
                        with Fig(fig, title='Temperature', size=mysize, time_series=False, proj=ccrs.PlateCarree()) as ax:

                            q = ax.scatter( G[lon], G[lat], c=G[T], s=todo_config_markersize, marker='o', edgecolor='black', linewidth=0.5,
                                           cmap='Spectral_r', transform=ccrs.PlateCarree())
                            ax.scatter( data[lon], data[lat], c='black', s=2, transform=ccrs.PlateCarree(), edgecolor='none', linewidth=0)
                            ax.set_xlim(data[lon].min()-lonpad, data[lon].max()+lonpad)
                            ax.set_ylim(data[lat].min()-latpad, data[lat].max()+latpad)
                            #ax.set_xticks(np.arange(data[lon].min(), data[lon].max(), lonpad))
                            ax.set_yticks(np.around(np.arange(bbox[2], bbox[3], (bbox[3]-bbox[2])/5),4))
                            ax.set_xticks(np.around(np.arange(bbox[0], bbox[1], (bbox[1]-bbox[0])/5),4))
                            ax.ticklabel_format(axis='both', style='plain', scilimits=[-10,6], useOffset=False, useMathText=True)

                            if not todo_config_map is None:
                                ax.add_image(osm_img, zoom)

                            cb = plt.colorbar(q, ax=ax, shrink=0.5, pad=0.03, aspect=30)
                            cb.set_label("Temperature (in $^\circ$C)", rotation=270, labelpad=20)
                            plt.grid(color='black', alpha=0.1)

                            if cc(config,'conversion','poi_plot','yes'):
                                for i, poi in poi_data.iterrows():
                                    ax.scatter( poi.lon, poi.lat, c='black', marker='+', s=100)

                    if cc(config, 'output', 'map_grid_tiles_humidity', 'yes'):
                        with Fig(fig, title='Relative air humidity', size=mysize, time_series=False, proj=ccrs.PlateCarree()) as ax:

                            q = ax.scatter( G[lon], G[lat], c=G[rh], s=todo_config_markersize, marker='o', edgecolor='black', linewidth=0.5,
                                           cmap='Spectral', transform=ccrs.PlateCarree())
                            ax.scatter( data[lon], data[lat], c='black', s=2, transform=ccrs.PlateCarree(), edgecolor='none', linewidth=0)
                            ax.set_xlim(data[lon].min()-lonpad, data[lon].max()+lonpad)
                            ax.set_ylim(data[lat].min()-latpad, data[lat].max()+latpad)
                            #ax.set_xticks(np.arange(data[lon].min(), data[lon].max(), lonpad))
                            ax.set_yticks(np.around(np.arange(bbox[2], bbox[3], (bbox[3]-bbox[2])/5),4))
                            ax.set_xticks(np.around(np.arange(bbox[0], bbox[1], (bbox[1]-bbox[0])/5),4))
                            ax.ticklabel_format(axis='both', style='plain', scilimits=[-10,6], useOffset=False, useMathText=True)

                            if not todo_config_map is None:
                                ax.add_image(osm_img, zoom)

                            cb = plt.colorbar(q, ax=ax, shrink=0.5, pad=0.03, aspect=30)
                            cb.set_label("Relative air humidity (in %)", rotation=270, labelpad=20)
                            plt.grid(color='black', alpha=0.1)

                            if cc(config,'conversion','poi_plot','yes'):
                                for i, poi in poi_data.iterrows():
                                    ax.scatter( poi.lon, poi.lat, c='black', marker='+', s=100)

                ## Plot interpolation (EXPERIMENTAL)
                elif var == 'map_interpolation':

                    var = sm
                    # TODO Those params should go to the config file
                    xi = np.arange(x.min()-xpad,x.max()+xpad, resolution)
                    yi = np.arange(y.min()-ypad,y.max()+ypad, resolution)
                    xi,yi = np.meshgrid(xi,yi)
                    # interpolate
                    zi = griddata((x,y),z,(xi,yi), method='linear') #, fill_value=0.3
                    # plot
                    with Fig(fig, title='Lin. Interpolation on %.0f m grid, %s' % (resolution, var), xlabel='Easting (in m)', ylabel='Northing (in m)', size=(10,10/(xrange/yrange)**2), time_series=False) as ax:
                        q = ax.contourf(xi, yi, zi, cmap='Spectral', vmin=smrange[0], vmax=smrange[1], levels=contour_levels, extend='both', corner_mask=False)
                        #q = plt.imshow(zi, cmap='Spectral', vmin=0.1, vmax=0.6, interpolation='gaussian', origin='lower', extent=(x.min()-3*xpad,x.max()+3*xpad, y.min()-3*ypad,y.max()+3*ypad))
                        ax.plot(x, y,'k.', ms=1.5)
                        #ax.set_xlabel('Easting [m]', fontsize=10)
                        #ax.set_ylabel('Northing [m]', fontsize=10)
                        #ax.set_title('Soil moisture linear interpolation')
                        cb = plt.colorbar(q, ax=ax, shrink=0.3, pad=0.03, aspect=30) #, ticks=np.arange(0,0.5,0.05)) #ticks=np.arange(np.ceil(np.nanmin(zi)*20)/20,np.nanmax(zi),0.05)
                        cb.set_label("vol. soil moisture (in m$^3$/m$^3$)", rotation=270, labelpad=20)
                        plt.grid(color='black', alpha=0.1)

                        if cc(config,'conversion','poi_plot','yes'):
                            for i, poi in poi_data.iterrows():
                                ax.scatter( poi.utmx-datax.x.min(), poi.utmy-datax.y.min(), c='black', marker='+', s=100)

                elif var in data.columns:
                    with Fig(fig, title=var, ylabel=var):
                        if not data[var].isnull().all():
                            plt.plot(data[var], ls='', marker='o', ms=5, markeredgewidth=0) #drawstyle='steps')
        # Done
        print('OK')
        file_save_msg(figure_name, "Figures", end="\n")

    # Make KML if lat/lon is provided
    if lat in data.columns and lon in data.columns:

        if yesno2bool(config['output']['make_KML_neutrons']):
            if not data[Nfinal].isnull().all():
                print('| Creating_KML for %s... ' % Nfinal)
                collim = xsplit(config['output']['KML_neutrons_range'])
                if len(collim)<=1: collim = None
                save_KML(data, Nfinal, lat, lon, alt=alt, file=output_basename+'-'+config['correction']['new_neutron_column'], format_str='%i', cmap='Spectral', collim=collim)#(3000,9000)) # long track: 5000,11000
                file_save_msg(output_basename+'-'+config['correction']['new_neutron_column']+'.kml', 'KML', end="\n")


#            if Nth:
#                print("\n"+'Creating_KML... ', end='')
#                save_KML(data, Nth, lat, lon, file=output_basename+'-'+config['correction']['new_neutron_column']+'_thermal', format_str='', cmap='Spectral') #, collim=(3000,9000)) # long track: 5000,11000
#                file_save_msg(output_basename+'-'+config['correction']['new_neutron_column']+'_thermal.kml', '', end='')

        if yesno2bool(config['output']['make_KML_sm']):
            if not data[sm].isnull().all():
                print('| Creating_KML for %s... ' % sm)
                collim = xsplit(config['output']['KML_sm_range'])
                if len(collim)<=1: collim = None
                save_KML(data, sm+'_%', lat, lon, alt=alt, file=output_basename+'-'+sm, format_str='%.0f%%', collim=collim)#(5,80))
                file_save_msg(output_basename+'-'+sm+'.kml', 'KML', end="\n")

        if yesno2bool(config['output']['make_KML_other']):
            other = config['output']['KML_other_column']
            if other in data.columns:
                if not data[other].isnull().all():
                    print('| Creating_KML for %s... ' % other)
                    collim = xsplit(config['output']['KML_other_range'])
                    if len(collim)<=1: collim = None
                    if yesno2bool(config['output']['KML_other_reverse']):
                        cmap = 'Spectral_r'
                    else:
                        cmap = 'Spectral'
                    format_str = '%'+config['output']['KML_other_format'] if config['output']['KML_other_format'] else ''
                    save_KML(data, other, lat, lon, alt=alt, file=output_basename+'-'+other, format_str=format_str, cmap=cmap, collim=collim)#(5,80))
                    file_save_msg(output_basename+'-'+other+'.kml', 'KML', end="\n")

                    # save also ID file
                    #save_KML(data, 'RecordNum', lat, lon, alt='Alt', file=output_basename+'-ID', format_str='%i')

     # Create the footprint geometries
    if cc(config,'output','make_footprint_polygons','yes'):
        print('| Creating footprint polygons...')
        #footprint_radius= gc(config, 'input','footprint_radius',dtype=float)
        #footprint_type=gc(config, 'input','footprint_type')
        #if footprint_type=='polygon':
        footprint.retrieve_footprints(data, footprint_radius = data.footprint_radius.mean(),
                                      output=os.path.join(output_basename+'_footprints.geojson'))
        file_save_msg(os.path.join(output_basename+'_footprints.geojson'), 'Polygons', end="\n")

    plt.close('all')
    print("= Done.")
    return(data)

# Direct call (instead of import)

if __name__ == "__main__":
    param2 = ''
    if len(sys.argv) > 1:
        main(sys.argv.pop(1))
        if len(sys.argv) > 1:
            param2 = sys.argv.pop(1)
    else:
        main()

    if param2 != '--nonstop':
        print("Press any key to close.", end='')
        input()
